/****************************************************************************
 *
 *                                 M U E S L I   v 1.0
 *
 *
 *     Copyright 2016 IMDEA Materials Institute, Getafe, Madrid, Spain
 *     Contact: muesli.materials@imdea.org
 *     Author: Ignacio Romero (ignacio.romero@imdea.org)
 *
 *     This file is part of MUESLI.
 *
 *     MUESLI is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     MUESLI is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with MUESLI.  If not, see <http://www.gnu.org/licenses/>.
 *
*****************************************************************************/



#include "finitestrain.h"

using namespace muesli;


finiteStrainMaterial :: finiteStrainMaterial()
: material(),
bulk(0.0),
cp(0.0), cs(0.0), rho(0.0)
{}




finiteStrainMaterial :: finiteStrainMaterial(const std::string& name,
                                                     const materialProperties& cl)
:
material(name, cl),
bulk(0.0),
cp(0.0), cs(0.0), rho(0.0)
{
    muesli::assignValue(cl, "density", rho);
}




bool finiteStrainMaterial :: check() const
{
    std::ostream& mlog = material::getLogger();

    if (cp == 0 && cs == 0)
    {
        mlog << "Error in finiteStrain material. No stiffness defined";
        return false;
    }
    return true;
}




double finiteStrainMaterial :: density() const
{
    return rho;
}




double finiteStrainMaterial :: getProperty(const propertyName p) const
{
    double ret=0.0;
    std::ostream& mlog = material::getLogger();


    // scan all the possible data
    switch(p)
    {
        case PR_BULK:   ret = bulk;		break;
        case PR_CP:		ret = cp;		break;
        case PR_CS:		ret = cs;		break;

        default:
            mlog << "property not defined";
    }
    return ret;
}




double finiteStrainMaterial :: waveVelocity() const
{
    return cp;
}




finiteStrainMP :: finiteStrainMP(const finiteStrainMaterial& m) :
theFiniteStrainMaterial(m)
{}





void finiteStrainMP :: CauchyStressVector(double S[6]) const
{
    istensor sigma;
    CauchyStress(sigma);
    muesli::tensorToVector(sigma, S);
}





void finiteStrainMP :: commitCurrentState()
{
    tn = tc;
    Fn = Fc;
    Jn = Jc;
}




void finiteStrainMP :: contractWithMaterialTangent(const ivector& v1, const ivector& v2, itensor& T) const
{
    itensor4 c;
    convectedTangent(c);

    for (unsigned i=0; i<3; i++)
        for (unsigned j=0; j<3; j++)
            for (unsigned k=0; k<3; k++)
                for (unsigned l=0; l<3; l++)
                {
                    T(i,j) += c(i,j,k,l)*v1(j)*v2(l);
                }
}




void finiteStrainMP :: contractWithSpatialTangent(const ivector& v1, const ivector& v2, itensor& T) const
{
    itensor4 c;
    spatialTangent(c);

    for (unsigned i=0; i<3; i++)
        for (unsigned j=0; j<3; j++)
            for (unsigned k=0; k<3; k++)
                for (unsigned l=0; l<3; l++)
                {
                    T(i,j) += c(i,j,k,l)*v1(j)*v2(l);
                }
}




void finiteStrainMP :: contractWithAllTangents(const ivector &v1,
                                                       const ivector& v2,
                                                       itensor&  Tdev,
                                                       istensor& Tmixed,
                                                       double&   Tvol) const
{
    contractWithDeviatoricTangent(v1, v2, Tdev);
    contractWithMixedTangent(Tmixed);
    Tvol = volumetricStiffness();
}





/*
 * Given the elasticity material C_{ABCD} = 2 d S_{AB} / d C_{CD} = 4 d^2 W(C) / d C^2,
 * and two material vectors V, W, obtain the (non-symmetric) tensor
 *
 *   T_{AB} = C_{ABCD} V_B W_D
 *
 *
 *  Although in some cases there might be a faster way, we reuse the function contractWithSpatialTangent
 *  through the expression:
 *
 *     C[ V, W] = J F^{-1} c[F^{-T]V, F^{-T}W] F^{-T}
 *
 */




void finiteStrainMP :: convectedTangentMatrix(double C[6][6]) const
{
    itensor4 ct;
    convectedTangent(ct);
    muesli::tensorToMatrix(ct, C);
}





double finiteStrainMP :: energyDissipationPotential() const
{
    double dt = tc - tn;
    return storedEnergy() + dt * dissipatedEnergy();
}




void finiteStrainMP :: energyMomentumTensor(const itensor &F, itensor &T) const
{
    itensor P;
    firstPiolaKirchhoffStress(P);

    T = (-1.0)*F.transpose()*P + storedEnergy()*itensor::identity();
}




void finiteStrainMP :: firstPiolaKirchhoffStress(itensor &P) const
{
    istensor sigma;
    CauchyStress(sigma);
    P = Jc * sigma * Fc.inverse().transpose();
}




materialState finiteStrainMP :: getConvergedState() const
{
    materialState state;

    state.theTime = tn;
    state.theDouble.push_back(Jn);
    state.theTensor.push_back(Fn);

    return state;
}




materialState finiteStrainMP :: getCurrentState() const
{
    materialState state;

    state.theTime = tc;
    state.theDouble.push_back(Jc);
    state.theTensor.push_back(Fc);

    return state;
}




void finiteStrainMP :: KirchhoffStress(istensor &tau) const
{
    istensor sigma;
    tau = sigma * Jc;
}




void finiteStrainMP :: KirchhoffStressVector(double tauv[6]) const
{
    istensor tau;
    KirchhoffStress(tau);
    muesli::tensorToVector(tau, tauv);
}




void finiteStrainMP :: materialTangentMatrix(double C[6][6]) const
{
    itensor4 mt;
    materialTangent(mt);
    muesli::tensorToMatrix(mt, C);
}




void finiteStrainMP :: materialTangentTimesSymmetricTensor(const istensor &M, istensor &CM) const
{
    itensor4 C;
    convectedTangent(C);

    for (unsigned i=0; i<3; i++)
        for (unsigned j=0; j<3; j++)
            for (unsigned k=0; k<3; k++)
                for (unsigned l=0; l<3; l++)
                {
                    CM(i,j) += C(i,j,k,l)*M(k,l);
                }
}




void finiteStrainMP :: resetCurrentState()
{
    tc = tn;
    Fc = Fn;
    Jc = Jn;
}




void finiteStrainMP :: secondPiolaKirchhoffStress(istensor& S) const
{
    istensor sigma;
    CauchyStress(sigma);
    S =  Jc * istensor::FSFt(Fc.inverse(), sigma);
}




void  finiteStrainMP :: secondPiolaKirchhoffStressVector(double Sv[6]) const
{
    istensor S;
    secondPiolaKirchhoffStress(S);
    muesli::tensorToVector(S, Sv);
}




void finiteStrainMP :: spatialTangentMatrix(double c[6][6]) const
{
    itensor4 st;
    spatialTangent(st);
    muesli::tensorToMatrix(st, c);
}




double finiteStrainMP :: stress(const double stretch) const
{
    std::ostream& mlog = material::getLogger();
    mlog << "\n Uniaxial stretch/stress relation not implemented for this material" << std::endl;
    return 0.0;
}




double finiteStrainMP :: stiffness(const double stretch) const
{
    std::ostream& mlog = material::getLogger();
    mlog << "\n Uniaxial stiffness not implemented for this material" << std::endl;
    return 0.0;
}




double finiteStrainMP :: storedEnergy(const double stretch) const
{
    std::ostream& mlog = material::getLogger();
    mlog << "\n Uniaxial stored energy function not implemented for this material" << std::endl;
    return 0.0;
}




void finiteStrainMP :: updateCurrentState(const double theTime, const itensor& F)
{
    tc = theTime;
    Fc = F;
    Jc = F.determinant();
}




double finiteStrainMP :: volumetricStiffness() const
{
    itensor4 tg;
    spatialTangent(tg);

    double vs = 0.0;
    for (unsigned i=0; i<3; i++)
    {
        for (unsigned j=0; j<3; j++)
            vs += 1.0/9.0 * tg(i,i,j,j);
    }
    return vs;
}




bool finiteStrainMP :: testImplementation(std::ostream& of, const bool testDE, const bool testDDE) const
{
    const unsigned voigt[2][6] = { {0,1,2,1,2,0}, {0,1,2,2,0,1}};
    bool isok = true;

    // set a random update in the material
    itensor F;
    F.setRandom();
    if (F.determinant() < 0.0) F *= -1.0;
    const_cast<finiteStrainMP*>(this)->updateCurrentState(0.0, F);
    const_cast<finiteStrainMP*>(this)->commitCurrentState();

    double tn1 = muesli::randomUniform(0.1,1.0);
    F.setRandom();
    if (F.determinant() < 0.0) F *= -1.0;
    const_cast<finiteStrainMP*>(this)->updateCurrentState(tn1, F);

    // programmed tangent of elasticities
    itensor4 tg;
    materialTangent(tg);

    itensor P;
    firstPiolaKirchhoffStress(P);
    const_cast<finiteStrainMP*>(this)->resetCurrentState();

    // material tangent A_{iAjB} = d (P_iA) / d F_jB
    itensor4 A;
    itensor dP, Pp1, Pp2, Pm1, Pm2;

    // numerical differentiation stress
    itensor numP;
    numP.setZero();
    const double   inc = 1.0e-4;

    for (size_t i=0; i<3; i++)
    {
        for (size_t j=0; j<3; j++)
        {
            double original = F(i,j);

            F(i,j) = original + inc;
            const_cast<finiteStrainMP*>(this)->updateCurrentState(tn1, F);
            double Wp1 = energyDissipationPotential();
            firstPiolaKirchhoffStress(Pp1);

            F(i,j) = original + 2.0*inc;
            const_cast<finiteStrainMP*>(this)->updateCurrentState(tn1, F);
            double Wp2 = energyDissipationPotential();
            firstPiolaKirchhoffStress(Pp2);

            F(i,j) = original - inc;
            const_cast<finiteStrainMP*>(this)->updateCurrentState(tn1, F);
            double Wm1 = energyDissipationPotential();
            firstPiolaKirchhoffStress(Pm1);

            F(i,j) = original - 2.0*inc;
            const_cast<finiteStrainMP*>(this)->updateCurrentState(tn1, F);
            double Wm2 = energyDissipationPotential();
            firstPiolaKirchhoffStress(Pm2);

            // fourth order approximation of the derivative
            numP(i,j) = (-Wp2 + 8.0*Wp1 - 8.0*Wm1 + Wm2)/(12.0*inc);

            // derivative of PK stress
            dP = (-Pp2 + 8.0*Pp1 - 8.0*Pm1 + Pm2)/(12.0*inc);
            for (unsigned k=0; k<3; k++)
                for (unsigned l=0; l<3; l++)
                    A(k,l,i,j) = dP(k,l);

            F(i,j) = original;
            const_cast<finiteStrainMP*>(this)->updateCurrentState(tn1, F);
        }
    }



    // compare DEnergy with the derivative of Energy
    if (testDE)
    {
        itensor errorP = numP - P;
        isok = (errorP.norm()/P.norm() < 1e-4);
        of << "\n   1. Comparing P with derivative of Weff.";
        if (isok)
        {
            of << " Test passed.";
        }
        else
        {
            of << "\n Relative error in DE computation %e. Test failed." << errorP.norm()/P.norm();
            of << "\n " << P;
            of << "\n " << numP;
        }
    }
    else
    {
        of << "\n   1. Comparing P with derivative of Weff ::: not run for this material";

    }


    // test the consistency of the stress tensors sigma and P
    {
        istensor sigma, S;
        CauchyStress(sigma);
        secondPiolaKirchhoffStress(S);

        itensor P1 = F.determinant() * sigma * F.inverse().transpose();
        itensor P2 = F*S;

        itensor errorP1 = P1 - P;
        itensor errorP2 = P2 - P;
        double  error = errorP1.norm() + errorP2.norm();
        isok = error/P.norm() < 1e-4;
        of << "\n   2. Checking consistency of the 1PK, 2PK, and Cauchy stresses.";
        if (isok)
        {
            of << " Test passed.";
        }
        else
        {
            of << "\n Test failed. Relative error" << error/P.norm();
            of << "\n " << P;
            of << "\n " << P1;
            of << "\n " << P2;
            of << "\n " << numP;
        }
    }


    // test tangent as derivative of the stress
    if (testDDE)
    {
        // numeric material tangent
        itensor4 nTg;
        nTg.setZero();

        // transform A to get the material tangent
        itensor  J  = F.inverse();
        istensor C  = istensor::tensorTransposedTimesTensor(F);
        istensor Ci = C.inverse();
        istensor S;
        secondPiolaKirchhoffStress(S);
        for (unsigned a=0; a<3; a++)
            for (unsigned b=0; b<3; b++)
                for (unsigned c=0; c<3; c++)
                    for (unsigned d=0; d<3; d++)
                    {
                        nTg(c,a,d,b) = - S(a,b)*Ci(c,d);

                        for (unsigned i=0; i<3; i++)
                            for (unsigned j=0; j<3; j++)
                                    nTg(c,a,d,b) += J(c,i)*A(i,a,j,b)*J(d,j);
                    }

        // relative
        double error = 0.0;
        double norm = 0.0;
        for (unsigned i=0; i<3; i++)
            for (unsigned j=0; j<3; j++)
                for (unsigned k=0; k<3; k++)
                    for (unsigned l=0; l<3; l++)
                    {
                        error += pow(nTg(i,j,k,l)-tg(i,j,k,l),2);
                        norm  += pow(tg(i,j,k,l),2);
                    }

        error = sqrt(error);
        norm = sqrt(norm);
        isok = (error/norm < 1e-4);

        of << "\n   3. Comparing material tangent with derivative of stress.";
        if (isok)
        {
            of << " Test passed.";
        }
        else
        {
            of << "\n      Test failed.";
            of << "\n      Relative error in DStress computation: " <<  error/norm;
        }
    }


    // Tests the function materialtangenttimes symmetric tensor and the
    // tangent contracted functions
    if ((true))
    {
        istensor Csym;
        istensor sym;  sym.setZero();
        ivector r1; r1.setRandom();
        ivector r2; r2.setRandom();
        sym.addSymmetrizedDyadic(r1, r2);
        materialTangentTimesSymmetricTensor(sym, Csym);

        istensor refCsym;   refCsym.setZero();
        itensor refCuv;     refCuv.setZero();
        for (unsigned i=0; i<3; i++)
            for (unsigned j=0; j<3; j++)
                for (unsigned k=0; k<3; k++)
                    for (unsigned l=0; l<3; l++)
                    {
                        // tangent times symmetric tensor
                        refCsym(i,j) += tg(i,j,k,l)*sym(k,l);

                        // inner contraction of tangent with two vectors
                        refCuv(i,k)  += tg(i,j,k,l)*r1(j)*r2(l);
                    }


        // relative error less than 0.01%
        istensor error = refCsym - Csym;
        isok = (error.norm()/Csym.norm() < 1e-4);
        of << "\n   4. Checking material tangent times symmetric tensor.";
        if (isok)
        {
            of << " Test passed.";
        }
        else
        {
            of << "\n   Test failed.";
            of << "\n   Relative error: " << error.norm()/Csym.norm();
            of << "\n   Csym \n" << Csym;
            of << "\n   Reference Csym:\n" << refCsym;
        }


        // relative error less than 0.01%
        itensor Cuv;
        contractWithMaterialTangent(r1, r2, Cuv);
        itensor Cuverror = refCuv - Cuv;
        isok = (Cuverror.norm()/Cuv.norm() < 1e-4);
        of << "\n   5. Checking contract material tangent.";
        if (isok)
        {
            of << " Test passed.";
        }
        else
        {
            of << "\n   Test failed.";
            of << "\n   Relative error: " << Cuverror.norm()/Cuv.norm();
            of << "\n   Cuv \n" << Cuv;
            of << "\n   Reference Cuv: \n" << refCuv;
        }

    }


    {
        // check voigt maps
        istensor sigma;
        CauchyStress(sigma);
        double S[6];
        CauchyStressVector(S);

        double error = 0.0;
        for (unsigned i=0; i<6; i++)
        {
            error += pow( S[i] - sigma( voigt[0][i] , voigt[1][i] ), 2);
        }

        // relative error less than 0.01%
        isok = (error/sigma.norm() < 1e-4);

        of << "\n   6. Comparing 2PK stress tensor and Voigt 2PK stress.";
        if (isok)
        {
            of << " Test passed.";
        }
        else
        {
            of << "\n      Test failed.";
            of << "\n      Relative error in stress comparison: " << error/sigma.norm();
        }
    }


    {
        double cm[6][6];
        materialTangentMatrix(cm);

        double error = 0.0;
        double norm  = 0.0;
        for (unsigned i=0; i<6; i++)
        {
            error += pow(cm[i][i] - tg( voigt[0][i], voigt[1][i], voigt[0][i], voigt[1][i] ),2);
            norm  += pow(cm[i][i],2);
            for (unsigned j=i+1; j<6; j++)
            {
                error += pow(cm[i][j] - tg( voigt[0][i], voigt[1][i], voigt[0][j], voigt[1][j] ),2);
                error += pow(cm[j][i] - tg( voigt[0][i], voigt[1][i], voigt[0][j], voigt[1][j] ),2);
                norm  += pow(cm[i][j],2);
                norm  += pow(cm[j][i],2);
            }
        }
        error = sqrt(error);
        norm  = sqrt(norm);


        // relative error less than 0.01%
        isok = (error/norm < 1e-4);
        of << "\n   7. Comparing material tangent tensor and Voigt matrix.";
        if (isok)
        {
            of << " Test passed.";
        }
        else
        {
            of << "\n      Test failed.";
            of << "\n      Relative error: " << error/norm;
        }
    }


    return isok;
}


