/****************************************************************************
 *
 *                                 M U E S L I   v 1.0
 *
 *
 *     Copyright 2016 IMDEA Materials Institute, Getafe, Madrid, Spain
 *     Contact: muesli.materials@imdea.org
 *     Author: Ignacio Romero (ignacio.romero@imdea.org)
 *
 *     This file is part of MUESLI.
 *
 *     MUESLI is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     MUESLI is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with MUESLI.  If not, see <http://www.gnu.org/licenses/>.
 *
*****************************************************************************/



#include "interface_abaqus.h"
#include "muesli/muesli.h"

using namespace muesli;

// matlabel:label referring to material type
// PROPS:   Material properties
// NPROPS:  Number of material properties
// TIME:    Current step time tn1, current accumulated time (2 values)
// DTIME:   Time step increment Delta t
// PNEWDT:  Proposed new time step (ratio)
// STRAN:   Strain at tn (11,22,33,12,13,23)
// DSTRAN:  Strain increment (11,22,33,12,13,23)
// DFGRD0:  Deformation gradient at time tn
// DFGRD1:  Deformation gradient at current time tn+1
// TEMP:    Temperature at time tn
// DTEMP:   Temperature increment from tn to tn+1
// STRESS:  Current stress (11,22,33,12,13,23)
// STATEV:  state variables in the last converged solution
// NSTATEV: Number of state variables
// SSE:     Elastic energy
// SPD:     Dissipated energy
// KSTEP:   Step number
// KINC:    Increment number, within the current step
//
int interface_abaqus_(char   CMNAME[80],
                      int*    matlabel,
                      int*    NOEL,
                      int*    NPT,
                      double* COORDS,
                      double* PROPS,
                      int*    NPROPS,
                      double* CELENT,
                      double* TIME,
                      double* DTIME,
                      double* PNEWDT,
                      double* STRAN,
                      double* DSTRAN,
                      double  DFGRD0[3][3],
                      double  DFGRD1[3][3],
                      double* TEMP,
                      double* DTEMP,
                      double* STRESS,
                      int*    NTENS,
                      double  DDSDDE[6][6],
                      double* STATEV,
                      int*    NSTATEV,
                      double* SSE,
                      double* SPD,
                      int*    KSTEP,
                      int*    KINC)
{

    switch (*matlabel)
    {

        case 1:
        {
            // linear elastic material
            //
            // matproperties[0] : E
            // matproperties[1] : nu
            const double E     = PROPS[0];
            const double nu    = PROPS[1];
            const double rho   = 1.0;

            elasticMaterial *theMaterial = new elasticMaterial(E, nu, rho);
            elasticMP       *thePoint    = dynamic_cast<elasticMP*>(theMaterial->createMaterialPoint());

            // set converged and current states
            istensor epsc(    STRAN[0]+DSTRAN[0],      STRAN[1]+DSTRAN[1],      STRAN[2]+DSTRAN[2],
                          .5*(STRAN[5]+DSTRAN[5]), .5*(STRAN[4]+DSTRAN[4]), .5*(STRAN[3]+DSTRAN[3]));

            double timec(TIME[0]);
            thePoint->updateCurrentState(timec, epsc);
            thePoint->commitCurrentState();

            // compute all the required fields
            double stressMuesli[6];
            thePoint->stressVector(stressMuesli);

            STRESS[0] = stressMuesli[0];
            STRESS[1] = stressMuesli[1];
            STRESS[2] = stressMuesli[2];
            STRESS[3] = stressMuesli[5];
            STRESS[4] = stressMuesli[4];
            STRESS[5] = stressMuesli[3];

            // tangent
            itensor4 tg;
            thePoint->tangentTensor(tg);

            const unsigned voigt[2][6] = { {0,1,2,0,2,1}, {0,1,2,1,0,2}};
            for (unsigned i=0; i<6; i++)
            {
                for (unsigned j=0; j<6; j++)
                {
                    DDSDDE[i][j] = tg(voigt[0][i], voigt[1][i], voigt[0][j], voigt[1][j]);
                }
            }

            delete thePoint;
            delete theMaterial;

        }
            break;

        case 2:
        {

            // small strain elastoplastic material
            //
            // material constants
            //
            const double E     = PROPS[0];   // Young modulus
            const double nu    = PROPS[1];   // Poisson ratio
            const double Y0    = PROPS[2];   // yield stress
            const double Hi    = PROPS[3];   // isotropic hardening modulus
            const double Hk    = PROPS[4];   // kinematic hardening modulus
            const double rho   = 1.0;        // density

            // create material and material point
            splasticMaterial *theMaterial = new splasticMaterial(E, nu, rho, Hi, Hk, Y0, "mises");
            splasticMP       *thePoint    = dynamic_cast<splasticMP*>(theMaterial->createMaterialPoint());

            // set converged states
            istensor epsn( STRAN[0], STRAN[1], STRAN[2], .5*STRAN[5], .5*STRAN[4], .5*STRAN[3]);
            istensor ep_n;
            double   dg_n;
            double   xi_n;
            istensor Xi_n;

            double tf(TIME[1]);
            if ( tf == 0.0 )
            {
                ep_n.setZero();
                dg_n = 0.0;
                xi_n = 0.0;
                Xi_n.setZero();
            }
            else
            {
                istensor ep_t( STATEV[0], STATEV[1], STATEV[2], STATEV[3], STATEV[4], STATEV[5]);
                ep_n = ep_t;
                dg_n = STATEV[6];
                xi_n = STATEV[7];
                istensor Xi_t( STATEV[8], STATEV[9], STATEV[10], STATEV[11], STATEV[12], STATEV[13]);
                Xi_n = Xi_t;
            }

            // set current states
            istensor epsc(    STRAN[0]+DSTRAN[0],      STRAN[1]+DSTRAN[1],      STRAN[2]+DSTRAN[2],
                          .5*(STRAN[5]+DSTRAN[5]), .5*(STRAN[4]+DSTRAN[4]), .5*(STRAN[3]+DSTRAN[3]));

            double timen(TIME[0]), timec(TIME[0]+*DTIME);

            thePoint->setConvergedState(timen, epsn, ep_n, xi_n, Xi_n);
            thePoint->updateCurrentState(timec, epsc);

            // compute all the required fields
            double stressMuesli[6];
            thePoint->stressVector(stressMuesli);

            STRESS[0] = stressMuesli[0];
            STRESS[1] = stressMuesli[1];
            STRESS[2] = stressMuesli[2];
            STRESS[3] = stressMuesli[5];
            STRESS[4] = stressMuesli[4];
            STRESS[5] = stressMuesli[3];

            // tangent
            itensor4 tg;
            thePoint->tangentTensor(tg);

            const unsigned voigt[2][6] = { {0,1,2,0,2,1}, {0,1,2,1,0,2}};
            for (unsigned i=0; i<6; i++)
            {
                for (unsigned j=0; j<6; j++)
                {
                    DDSDDE[i][j] = tg( voigt[0][i], voigt[1][i], voigt[0][j], voigt[1][j]);
                }
            }

            // recover and store state variables
            double xi_c;
            istensor ep_c, Xi_c;

            materialState state_c = thePoint->getCurrentState();

            xi_c = state_c.theDouble[1];
            ep_c = state_c.theStensor[1];
            Xi_c = state_c.theStensor[2];

            // store the updated state variables
            STATEV[0] = ep_c(0,0);
            STATEV[1] = ep_c(1,1);
            STATEV[2] = ep_c(2,2);
            STATEV[3] = ep_c(1,2);
            STATEV[4] = ep_c(2,0);
            STATEV[5] = ep_c(0,1);
            //STATEV[6] = dg_c;
            STATEV[7] = xi_c;
            STATEV[8] = Xi_c(0,0);
            STATEV[9] = Xi_c(1,1);
            STATEV[10] = Xi_c(2,2);
            STATEV[11] = Xi_c(1,2);
            STATEV[12] = Xi_c(2,0);
            STATEV[13] = Xi_c(0,1);
            
            delete thePoint;
            delete theMaterial;
        }
            
        default:
            break;
    }
    
    
    return 0;
}


