/****************************************************************************
 *
 *                                 M U E S L I   v 1.0
 *
 *
 *     Copyright 2016 IMDEA Materials Institute, Getafe, Madrid, Spain
 *     Contact: muesli.materials@imdea.org
 *     Author: Ignacio Romero (ignacio.romero@imdea.org)
 *
 *     This file is part of MUESLI.
 *
 *     MUESLI is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     MUESLI is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with MUESLI.  If not, see <http://www.gnu.org/licenses/>.
 *
*****************************************************************************/



#include "elastic.h"
#include <cmath>

using namespace std;
using namespace muesli;


elasticMaterial :: elasticMaterial() :
  smallStrainMaterial(),
  E(0.0), nu(0.0), lambda(0.0), mu(0.0), bulk(0.0), cp(0.0), cs(0.0), rho(1.0)
{
}




elasticMaterial :: elasticMaterial(const double xE, const double xnu,
                                   const double xrho)
  :
  E(xE), nu(xnu), rho(xrho)
{
  lambda = E*nu/(1.0-2.0*nu)/(1.0+nu);
  mu     = E/2.0/(1.0+nu);
  bulk   = lambda + 2.0/3.0 * mu;

  if (rho > 0.0)
    {
      cp   = sqrt((lambda+2.0*mu)/rho);
      cs   = sqrt(2.0*mu/rho);
    }
}




elasticMaterial :: elasticMaterial(const std::string& name,
				   const materialProperties& cl)
  :
  smallStrainMaterial(name, cl),
  E(0.0), nu(0.0),
  lambda(0.0), mu(0.0), bulk(0.0),
  cp(0.0), cs(0.0),
  rho(0.0)
{
  muesli::assignValue(cl, "young",   E);
  muesli::assignValue(cl, "poisson", nu);
  muesli::assignValue(cl, "lambda",  lambda);
  muesli::assignValue(cl, "mu",      mu);
  muesli::assignValue(cl, "density", rho);
    
  // intialize rest mlog constants
  if (lambda*lambda + mu*mu + E*E == 0)
    std::cout << "\n Error in small strain elastic material. No stiffness defined";

  // E and nu have priority. If they are defined, define lambda and mu
  if (E*E > 0)
    {
      lambda = E*nu/(1.0-2.0*nu)/(1.0+nu);
      mu     = E/2.0/(1.0+nu);
    }
  else
    {
      nu = lambda / 2.0 / (lambda+mu);
      E  = mu*2.0*(1.0+nu);
    }

  // we set all the constants, so that later on all mlog them can be recovered fast
  bulk = lambda + 2.0/3.0 * mu;

  if (rho > 0.0)
    {
      cp   = sqrt((lambda+2.0*mu)/rho);
      cs   = sqrt(2.0*mu/rho);
    }
}




bool elasticMaterial :: check() const
{
  if (mu > 0 && lambda+2.0*mu > 0) return true;
  else return false;
}




/* an object mlog the type "elastic" creates a material point mlog type elastic. The
   material point holds information that is not part mlog the material itself but
   that is particular mlog the specific (physical) point.
*/
smallStrainMP* elasticMaterial :: createMaterialPoint() const
{
  smallStrainMP* mp = new elasticMP(*this);
  return mp;
}




double elasticMaterial :: density() const
{
  return rho;
}




// this function is much faster than the one with string property names, because
// avoids string comparisons. It should be used.
double elasticMaterial :: getProperty(const propertyName p) const
{
  double ret=0.0;

  switch (p)
    {
    case PR_LAMBDA:     ret = lambda;   break;
    case PR_MU:         ret = mu;       break;
    case PR_YOUNG:      ret = E;        break;
    case PR_POISSON:    ret = nu;       break;
    case PR_BULK:       ret = bulk;     break;
    case PR_CP:         ret = cp;       break;
    case PR_CS:         ret = cs;       break;

    default:
      std::cout << "\n Error in elasticMaterial. Property not defined";
    }
  return ret;
}




/* this function is always called once the material is defined, so apart from
   printing its information, we take the opportunity to clean up some mlog its
   data, in particular, setting all the possible constants
*/
void elasticMaterial :: print(std::ostream &of) const
{
  std::ostream& mlog = material::getLogger();

  mlog  << "\n   Small strain, elastic, isotropic material ";
  mlog  << "\n   Young modulus   E      : " << E;
  mlog  << "\n   Poisson ratio   nu     : " << nu;
  mlog  << "\n   Lame constants  lambda : " << lambda;
  mlog  << "\n                   mu     : " << mu;
  mlog  << "\n   Bulk modulus    k      : " << bulk;
  mlog  << "\n   Density                : " << rho;
  if (rho > 0.0)
    {
      mlog  << "\n   Wave velocities C_p    : " << cp;
      mlog  << "\n                   C_s    : " << cs;
    }
  mlog << std::flush;
}




void elasticMaterial :: setRandom()
{
  E      = muesli::randomUniform(1.0, 10000.0);
  nu     = muesli::randomUniform(0.05, 0.45);
  rho    = muesli::randomUniform(1.0, 100.0);
  lambda = E*nu/(1.0-2.0*nu)/(1.0+nu);
  mu     = E/2.0/(1.0+nu);
  cp     = sqrt((lambda+2.0*mu)/rho);
  cs     = sqrt(2.0*mu/rho);
  bulk   = lambda + 2.0/3.0 * mu;
}




bool elasticMaterial :: test(std::ostream &of)
{
  bool isok = true;
  setRandom();
  smallStrainMP* p = this->createMaterialPoint();

  isok = p->testImplementation(of);
  delete p;
  return isok;
}




double elasticMaterial :: waveVelocity() const
{
  return cp;
}




elasticMP :: elasticMP(const elasticMaterial &m) :
  muesli::smallStrainMP(m),
  theElasticMaterial(m)
{
}




elasticMP :: ~elasticMP()
{

}





/* Given the fourth order tensor mlog elasticities C, and two vectors v, w
 * compute the second order tensor T with components
 *   T_ij = C_ipjq v_p w_q
 *
 *  which, for linear isotropic materials is just
 *  T = lambda gradV otimes gradW + mu gradW otimes gradV + mu (gradV * gradW) One
 *
 *  Note that the result is not symmetric.
 */

void elasticMP :: contractWithTangent(const ivector &v1, const ivector &v2, itensor &T) const
{
  itensor A, B;

  T = itensor::identity();
  T *= v1.dot(v2) * theElasticMaterial.mu;

  A = itensor::dyadic(v1, v2);
  B = A.transpose();

  A *= theElasticMaterial.lambda;
  B *= theElasticMaterial.mu;

  T += A;
  T += B;
}




void elasticMP :: contractWithDeviatoricTangent(const ivector &v1, const ivector &v2, itensor &T) const
{
  const double&  mu = theElasticMaterial.mu;

  T = mu * v1.dot(v2) * itensor::identity()
    +   mu *              itensor::dyadic(v2, v1)
    -   2.0/3.0*mu *      itensor::dyadic(v1, v2);
}




double elasticMP :: deviatoricEnergy() const
{
  istensor e = istensor::deviatoricPart(eps_c);
  return theElasticMaterial.mu * e.contract(e);
}




void elasticMP :: deviatoricStress(istensor& s) const
{
  s = (2.0*theElasticMaterial.mu)* istensor::deviatoricPart(eps_c);
}




double elasticMP :: dissipatedEnergy() const
{
  return 0.0;
}




double elasticMP :: energyDissipationPotential() const
{
  return storedEnergy();
}




materialState elasticMP :: getConvergedState() const
{
  materialState st;
  st.theTime = time_n;
  st.theStensor.push_back(eps_n);

  return st;
}




materialState elasticMP :: getCurrentState() const
{
  materialState st;
  st.theTime = time_c;
  st.theStensor.push_back(eps_c);

  return st;
}




double elasticMP :: plasticSlip() const
{
  return 0.0;
}




double elasticMP :: pressure() const
{
  double p = -theElasticMaterial.bulk * eps_c.trace();
  return p;
}




void elasticMP :: setConvergedState(const double xtn, const istensor& strainn)
{
  time_n = xtn;
  eps_n = strainn;
}




double elasticMP :: storedEnergy() const
{
  double W;
  if (uniaxial)
    {
      W = 0.5*theElasticMaterial.E*eps_c(0,0)*eps_c(0,0);
    }
  else
    {
      const double  tr = eps_c.trace();
      W  =  theElasticMaterial.mu*eps_c.squaredNorm() + 0.5*theElasticMaterial.lambda*tr*tr;
    }
  return W;
}




void elasticMP :: stress(istensor& sigma) const
{
  sigma = (2.0*theElasticMaterial.mu)*eps_c + (theElasticMaterial.lambda*eps_c.trace()) * istensor::identity();
}




void elasticMP :: stressVector(double S[6]) const
{
  const double nu = theElasticMaterial.nu;
  const double k  = theElasticMaterial.E/(1.0+nu)/(1.0-2.0*nu);

  S[0] = k*(1.0-nu) * eps_c(0,0) + k*nu * eps_c(1,1) + k*nu * eps_c(2,2);
  S[1] = k*(1.0-nu) * eps_c(1,1) + k*nu * eps_c(2,2) + k*nu * eps_c(0,0);
  S[2] = k*(1.0-nu) * eps_c(2,2) + k*nu * eps_c(0,0) + k*nu * eps_c(1,1);
  S[3] = k*(1.0-2.0*nu)* eps_c(1,2);
  S[4] = k*(1.0-2.0*nu)* eps_c(2,0);
  S[5] = k*(1.0-2.0*nu)* eps_c(0,1);
}




void elasticMP :: tangentMatrix(double C[6][6]) const
{
  const double nu = theElasticMaterial.nu;
  const double k  = theElasticMaterial.E/(1.0+nu)/(1.0-2.0*nu);

  for (unsigned i=0; i<6; i++)
    for (unsigned j=0; j<6; j++)
      C[i][j] = 0.0;

  for (unsigned i=0; i<3; i++)
    {
      C[i][i]     = (1.0-nu)*k;
      C[i+3][i+3] = 0.5*(1.0-2.0*nu)*k;
      for (unsigned j=i+1; j<3; j++)
	C[i][j] = C[j][i] = nu*k;
    }
}




void elasticMP :: tangentTensor(itensor4& C) const
{
  C.setZero();

  for (unsigned i=0; i<3; i++)
    for (unsigned j=0; j<3; j++)
      for (unsigned k=0; k<3; k++)
	for (unsigned l=0; l<3; l++)
	  {
	    if (i==j && k==l) C(i,j,k,l) += theElasticMaterial.lambda;
	    if (i==k && j==l) C(i,j,k,l) += theElasticMaterial.mu;
	    if (i==l && j==k) C(i,j,k,l) += theElasticMaterial.mu;
	  }
}




double elasticMP :: uniaxialStiffness() const
{
  return theElasticMaterial.E;
}




double elasticMP :: uniaxialStress() const
{
  return theElasticMaterial.E * eps_c(0,0);
}




double elasticMP :: volumetricStiffness() const
{
  return theElasticMaterial.bulk;
}




double elasticMP :: volumetricEnergy() const
{
  double theta = eps_c.trace();
  double V     = 0.5*theElasticMaterial.bulk * theta * theta;
  return V;
}




