/****************************************************************************
 *
 *                                 M U E S L I   v 1.0
 *
 *
 *     Copyright 2016 IMDEA Materials Institute, Getafe, Madrid, Spain
 *     Contact: muesli.materials@imdea.org
 *     Author: Ignacio Romero (ignacio.romero@imdea.org)
 *
 *     This file is part of MUESLI.
 *
 *     MUESLI is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     MUESLI is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with MUESLI.  If not, see <http://www.gnu.org/licenses/>.
 *
*****************************************************************************/


#include <string.h>
#include <cmath>
#include "splastic.h"
#include <cassert>

#define PROJTOL     1e-8     // residual tolerance for return-to-yield-surface solutions
#define PROJMAXITER 25      // max num of allowed interations in return-to-yield solutions


/*
 * splastic.cpp,
 * infinitesimal kinematics visco-elastoplastic element
 * d. del pozo, i. romero, feb 2015
 *
 *
 *  1) von Mises Model with linear isotropic and kinematic hardenings
 *
 *     See "Numerical analysis and simulation of plasticity", by J.C. Simo,
 *     Handbook of numerical analysis vol VI
 *     table 18.2, p.257 and pages close to it
 *
 *     We = kappa/2 (theta)^2 + mu * ||e_elastic||^2
 *     Wp = Hiso/2 xi^2 + 2/3 Hkine/2 ||Xi||^2
 *     f(sigma, q, Q) = || dev(sigma) - Q || - sqrt(2/3) (Y0 - q)
 *
 *
 *  2) Drucker-Praguer Model
 *     linear isotropic hardening
 *
 *     The yield function depends on :
 *        - I: trace of sigma
 *        - J: square root of J2
 *        - q: conjugate stress of the internal isotropic variable
 *
 *        f(I,J,q) = J + alphac*I - (alphac+sqrt(3)/3) * (Y0 - q);
 *        q = -K'(xi),    K(xi) = Hiso/2 xi^2
 *
 *
 *  3) Tresca with linear isotropic hardening
 *
 *     We = kappa/2 (theta)^2 + mu * ||e_elastic||^2
 *     Wp = Hiso/2 xi^2
 *     f(sigma, q, Q) = sigma_I - sigma_III - (Y0 - q)
 *
 *
 *  4) 1D Model
 *
 *        See "Computational Inelasticity", Box 1.5, p.45
 *
 *        Q = Hkine * Xi(0,0);
 *        q = Hiso  * xi(0,0);
 *        f = sigma - Q - (Y0 - q);
 *
 */


using namespace std;
using namespace muesli;


splasticMaterial:: splasticMaterial() :
  smallStrainMaterial(),
  plasticityType("mises"),
  E(0.0), nu(0.0), bulk(0.0), cp(0.0), cs(0.0), lambda(0.0), mu(0.0), rho(0.0),
  Hiso(0.0), Hkine(0.0), Y0(0.0),
  alphac(0.0)
{
}


splasticMaterial :: splasticMaterial(const double xE,     const double xnu,
                                     const double xrho,   const double xHiso,
                                     const double xHkine, const double xY0,
                                     const char*  plasType)
  :
  E(xE), nu(xnu), bulk(0.0),
  cp(0.0), cs(0.0), lambda(0.0), mu(0.0), rho(xrho),
  Hiso(xHiso), Hkine(xHkine), Y0(xY0),
  alphac(0.0)
{
  plasticityType = std::string(plasType);

  lambda = E*nu/(1.0-2.0*nu)/(1.0+nu);
  mu     = E/2.0/(1.0+nu);
  bulk   = lambda + 2.0/3.0 * mu;

  if (rho > 0.0)
    {
      cp   = sqrt((lambda+2.0*mu)/rho);
      cs   = sqrt(2.0*mu/rho);
    }
}




splasticMaterial :: splasticMaterial(const double xE, const double xnu,
                                     const double xrho, const double xHiso,
                                     const double xHkine, const double xY0,
                                     const std::string xplst)
  :
  plasticityType(xplst),
  E(xE), nu(xnu), bulk(0.0), cp(0.0), cs(0.0),
  lambda(0.0), mu(0.0), rho(xrho),
  Hiso(xHiso), Hkine (xHkine), Y0(xY0), alphac(0.0)
{

  lambda = E*nu/(1.0-2.0*nu)/(1.0+nu);
  mu     = E/2.0/(1.0+nu);
  bulk   = lambda + 2.0/3.0 * mu;

  if (rho > 0.0)
    {
      cp   = sqrt((lambda+2.0*mu)/rho);
      cs   = sqrt(2.0*mu/rho);
    }
}





splasticMaterial :: splasticMaterial(const std::string& name,
                                     const materialProperties& cl)
  :
  smallStrainMaterial(name, cl),
  plasticityType("mises"),
  E(0.0), nu(0.0), bulk(0.0), cp(0.0), cs(0.0), lambda(0.0), mu(0.0), rho(0.0),
  Hiso(0.0), Hkine(0.0), Y0(0.0), alphac(0.0)
{
  muesli::assignValue(cl, "young",   E);
  muesli::assignValue(cl, "poisson", nu);
  muesli::assignValue(cl, "lambda",  lambda);
  muesli::assignValue(cl, "mu",      mu);
  muesli::assignValue(cl, "density", rho);
  muesli::assignValue(cl, "isotropich", Hiso);
  muesli::assignValue(cl, "kinematich", Hkine);
  muesli::assignValue(cl, "yieldstress", Y0);
  muesli::assignValue(cl, "alphac", alphac);
    
  if (cl.find("model drucker")   != cl.end()) plasticityType = "drucker";
  if (cl.find("model von_mises") != cl.end()) plasticityType = "mises";
  if (cl.find("model tresca") != cl.end()) plasticityType = "tresca";
    
  
  // intialize rest mlog constants
  if (lambda*lambda + mu*mu + E*E == 0)
    std::cout << "\n Error in small strain elastic material. No stiffness defined";
	
  // E and nu have priority. If they are defined, define lambda and mu
  if (E*E > 0)
    {
      lambda = E*nu/(1.0-2.0*nu)/(1.0+nu);
      mu     = E/2.0/(1.0+nu);
    }
  else
    {
      nu = lambda / 2.0 / (lambda+mu);
      E  = mu*2.0*(1.0+nu);
    }
	
  // we set all the constants, so that later on all mlog them can be recovered fast
  bulk = lambda + 2.0/3.0 * mu;

  if (rho > 0.0)
    {
      cp   = sqrt((lambda+2.0*mu)/rho);
      cs   = sqrt(2.0*mu/rho);
    }
}





bool splasticMaterial :: check() const
{
  if (mu > 0 && lambda+2.0*mu > 0) return true;
  else return false;
}




smallStrainMP* splasticMaterial :: createMaterialPoint() const
{
  smallStrainMP* mp = new splasticMP(*this);
  return mp;
}



double splasticMaterial :: density() const
{
  return rho;
}




// this function is much faster than the one with string property names, because
// avoids string comparisons. It should be used.
double splasticMaterial :: getProperty(const propertyName p) const
{
  double ret=0.0;
    
  switch (p)
    {
    case PR_LAMBDA:     ret = lambda;   break;
    case PR_MU:         ret = mu;       break;
    case PR_YOUNG:      ret = E;        break;
    case PR_POISSON:    ret = nu;       break;
    case PR_BULK:       ret = bulk;     break;
    case PR_CP:         ret = cp;       break;
    case PR_CS:         ret = cs;       break;
    case PR_YIELD:      ret = Y0;       break;
    case PR_ISOHARD:    ret = Hiso;     break;
    case PR_KINHARD:    ret = Hkine;    break;
            
    default:
      std::cout << "Error in elastoplasticMaterial. Property not defined";
    }
  return ret;
}


void splasticMaterial :: print(std::ostream &of) const
{
    
  if (plasticityType == "mises")
    {
      of  << "\n Elastoplastic material for small strain kinematics."
	  << "\n   von Mises yield criterion with linear isotropic and kinematic hardenings."
	  << "\n   Yield stress           : " << Y0
	  << "\n   Isotropic hardening    : " << Hiso
	  << "\n   Kinematic hardening    : " << Hkine;
    }
    
    
  else if (plasticityType == "tresca")
    {
      of  << "\n Elastoplastic material for small strain kinematics."
	  << "\n   Tresca yield criterion with linear isotropic and kinematic hardenings."
	  << "\n   Yield stress           : " << Y0
	  << "\n   Isotropic hardening    : " << Hiso
	  << "\n   Kinematic hardening    : " << Hkine;
    }


  else if (plasticityType == "drucker")
    {
      of  << "\n Drucker-Praguer yield criterion with linear isotropic hardenings."
	  << "\n Yield function --> F = J + alpha*I - (3.0*alpha+sqrt(3.0))/3.0*(Y0-q)"
	  << "\n J = sqrt(J2), I = trace(stresstensor)"
	  << "\n Conjugate force --> q = -K'(xi), K(xi) = 1/2 Hiso xi^2 "
	  << "\n Yield stress, tensile stress for pressure dependent models: Y0          : " << Y0
	  << "\n Drucker Prager's cone semiangle:                            alpha       : " << alphac
	  << "\n Isotropic hardening:                                        Hiso        : " << Hiso;
    }
    
    
  of  << "\n   Young modulus:  E      : " << E;
  of  << "\n   Poisson ratio:  nu     : " << nu;
  of  << "\n   Lame constants: lambda : " << lambda;
  of  << "\n   Shear modulus:  mu     : " << mu;
  of  << "\n   Bulk modulus:   k      : " << bulk;
  of  << "\n   Density                : " << rho;
    
  if (rho > 0.0)
    {
      of  << "\n   Wave velocities c_p    : " << cp;
      of  << "\n                   c_s    : " << cs;
    }
}




void splasticMaterial :: setRandom()
{
  E   = muesli::randomUniform(1.0, 10000.0);
  nu  = muesli::randomUniform(0.05, 0.45);
  rho = muesli::randomUniform(1.0, 100.0);
    
  lambda = E*nu/(1.0-2.0*nu)/(1.0+nu);
  mu     = E/2.0/(1.0+nu);
  cp     = sqrt((lambda+2.0*mu)/rho);
  cs     = sqrt(2.0*mu/rho);
  bulk   = lambda + 2.0/3.0*mu;
    
    
  int pt = muesli::discreteUniform(0, 1);
  if (pt == 0)
    {
        
      plasticityType = "mises";
      Y0             = E* muesli::randomUniform(0.5, 1.5);
      Hiso           = E*1e-4 * muesli::randomUniform(1.0, 2.0);
      Hkine          = E*1e-4 * muesli::randomUniform(1.0, 2.0);
    }
    
  else if (pt == 1)
    {
      plasticityType = "drucker";
        
      Y0            = E*1e-3* muesli::randomUniform(0.5, 1.5);
      alphac        = muesli::randomUniform(0.1,0.3 );
      Hiso          = E*1e-4 * muesli::randomUniform(1.0, 2.0);
      Hkine         = 0.0;
    }
}




bool splasticMaterial :: test(std::ostream  &of)
{
  setRandom();
  smallStrainMP* p = this->createMaterialPoint();
  p->setRandom();
  if      (this->plasticityType == "mises")   of << "\n   von Mises type";
  else if (this->plasticityType == "drucker") of << "\n   Drucker-Prager type";
  else if (this->plasticityType == "tresca") of << "\n    Tresca type";

  bool ok = p->testImplementation(of);
  delete p;
  return ok;
}




double splasticMaterial :: waveVelocity() const
{
  return cp;
}




splasticMP :: splasticMP(const splasticMaterial &m) :
  muesli::smallStrainMP(m),
  theElastoplasticMaterial(m),
  dg_n(0.0), xi_n(0.0),
  dg_c(0.0), xi_c(0.0)
{
  ep_n.setZero();
  ep_c.setZero();
  Xi_n.setZero();
  Xi_c.setZero();
}




void splasticMP :: commitCurrentState()
{
  smallStrainMP::commitCurrentState();
  dg_n   = dg_c;
  ep_n   = ep_c;
  xi_n   = xi_c;
  Xi_n   = Xi_c;
}




void splasticMP :: contractWithDeviatoricTangent(const ivector &v1, const ivector &v2, itensor &T) const
{
  const double mu  = theElastoplasticMaterial.mu;
  const double k   = theElastoplasticMaterial.bulk;
  const double Hiso = theElastoplasticMaterial.Hiso;


  if (theElastoplasticMaterial.plasticityType == "mises")
    {
      if (dg_c > 0.0)
        {
	  const double   Hkine    = theElastoplasticMaterial.Hkine;
	  const istensor strial   = 2.0*mu*(istensor::deviatoricPart(eps_c) - ep_n);
	  const istensor Qtrial   = -2.0/3.0*Hkine*Xi_n;
        
	  const istensor backs    = strial - Qtrial;
	  const double   betanorm = backs.norm();
	  const istensor n        = backs * (1.0/betanorm);
	  const double   zeta     = 1.0 - 2.0*mu*dg_c/betanorm;
	  const double   zbar     = 2.0*mu/(2.0*mu+2.0/3.0*(Hiso+Hkine)) - (1.0-zeta);
        
	  T = - 2.0/3.0*mu*zeta   * itensor::dyadic(v1, v2)
            +   mu*zeta*v1.dot(v2)  * itensor::identity()
            +   mu*zeta             * itensor::dyadic(v2,v1)
            -   2.0*mu*zbar         * itensor::dyadic(n*v1,n*v2);
        }
      else
        {
	  T = - 2.0/3.0*mu        * itensor::dyadic(v1, v2)
            +   mu*v1.dot(v2)       * itensor::identity()
            +   mu                  * itensor::dyadic(v2,v1);
        }
    }

  else if (theElastoplasticMaterial.plasticityType == "drucker")
    {
      if (dg_c == 0.0)
        {
	  T = - 2.0/3.0*mu        * itensor::dyadic(v1, v2)
            +   mu*v1.dot(v2)       * itensor::identity()
            +   mu                  * itensor::dyadic(v2,v1);
        }
      else
        {
	  const istensor edtrial     = istensor::deviatoricPart(eps_c-ep_n);//deviatoric elastic trial strain
	  const istensor evoltrial   = eps_c-ep_n-edtrial;//trial elastic volumetric strain
	  const istensor D           = edtrial * (1.0/edtrial.norm());
	  const double   alphac      = theElastoplasticMaterial.alphac;
	  const double   normed      = sqrt(2.0)*edtrial.norm();
	  const double   beta        = (3.0*alphac+sqrt(3.0))/3.0;
	  const double   A           = 1.0/(mu+9.0*k*alphac*alphac+beta*beta*Hiso);
	  const istensor s_trial     = 2.0 * mu * edtrial;
	  const double   J_trial     = sqrt(s_trial.J2());
            
	  bool isInCone = J_trial-mu*dg_c >= 0.0;
	  if (isInCone)//case within cone's definition
            {
	      T = mu*(1.0-dg_c/normed)* v1.dot(v2)* itensor::identity()
		+ mu*(1.0-dg_c/normed)*itensor::dyadic(v2,v1)
		- 2.0*mu*(1.0-dg_c/normed)/3.0*itensor::dyadic(v1, v2)
		+ 2.0*mu*(dg_c/normed)*itensor::dyadic(D*v1, D*v2)
		- 2.0*mu*mu*A*itensor::dyadic(D*v1, D*v2);
            }
	  else //case beyond the apex
            {
	      istensor Q; Q.setZero();
                
	      T = Q;
            }
        }
    }
}




void splasticMP :: contractWithTangent(const ivector &v1, const ivector &v2, itensor &T) const
{
  const double mu  = theElastoplasticMaterial.mu;
  const double k   = theElastoplasticMaterial.bulk;
  const double Hkine = theElastoplasticMaterial.Hkine;
  const double Hiso = theElastoplasticMaterial.Hiso;
    
    
    
  if (theElastoplasticMaterial.plasticityType == "mises")
    {
      if (dg_c == 0.0)
        {
	  T = (k - 2.0/3.0*mu) * itensor::dyadic(v1, v2)
            + mu*v1.dot(v2)      * itensor::identity()
            + mu                 * itensor::dyadic(v2,v1);
        }
      else
        {
	  const istensor strial = 2.0*mu*(istensor::deviatoricPart(eps_c) - ep_n);
	  const istensor Qtrial = -2.0/3.0*Hkine*Xi_n;
	  const istensor backs  = strial - Qtrial;
	  const istensor n      = backs * (1.0/backs.norm());
	  const double zeta     = 1.0 - 2.0*mu*dg_c/backs.norm();
	  const double zbar     = 2.0*mu / (2.0*mu+2.0/3.0*(Hiso+Hkine)) - (1.0-zeta);

	  T = (k - 2.0/3.0*mu*zeta)  * itensor::dyadic(v1, v2)
            + (mu*zeta*v1.dot(v2)) * itensor::identity()
            + (mu*zeta)            * itensor::dyadic(v2,v1)
            - (2.0*mu*zbar)        * itensor::dyadic(n*v1,n*v2);
        }
    }
    
    
  else if (theElastoplasticMaterial.plasticityType == "drucker")
    {
      // Elastic part
      if (dg_c == 0.0)
        {
	  T = (k - 2.0/3.0*mu) * itensor::dyadic(v1, v2)
            + mu*v1.dot(v2)   * itensor::identity()
            + mu              * itensor::dyadic(v2,v1);
        }
      // Elastoplastic part
      else
        {
	  const istensor edtrial     = istensor::deviatoricPart(eps_c-ep_n);//deviatoric elastic trial strain
	  const istensor D           = edtrial * (1.0/edtrial.norm());
	  const double   alphac      = theElastoplasticMaterial.alphac;
	  const double   normed      = sqrt(2.0)*edtrial.norm();
	  const double   root2       = sqrt(2.0);
	  const double   beta        = (3.0*alphac+sqrt(3.0))/3.0;
	  const double   A           = 1.0/(mu+9.0*k*alphac*alphac+beta*beta*Hiso);
	  const istensor s_trial     = 2.0 * mu * edtrial;
	  const double   J_trial     = sqrt(s_trial.J2());
            
	  bool isInCone = J_trial-mu*dg_c >= 0.0;
	  if (isInCone)//case within cone's definition
	    {
	      T =
                
		//Deviatoric Part
		mu*(1.0-dg_c/normed)* v1.dot(v2)* itensor::identity()
		+ mu*(1.0-dg_c/normed)*itensor::dyadic(v2,v1)
		- 2.0/3.0*mu*(1.0-dg_c/normed)*itensor::dyadic(v1, v2)
		+ 2.0*mu*(dg_c/normed)*itensor::dyadic(D*v1, D*v2)
		- 2.0*mu*mu*A*itensor::dyadic(D*v1, D*v2)
		- A*root2*mu*3.0*k*alphac*itensor::dyadic(D*v1, v2)

		//Volumetric Part
		+ k*itensor::dyadic(v1, v2)
		- A*root2*mu*3.0*k*alphac*itensor::dyadic(v1, D*v2)
		- alphac*alphac*9.0*k*k*A*itensor::dyadic(v1, v2);
	    }
	  else //case beyond the apex
            {
	      const double   Q = 1.0/(3.0*k*alphac+beta*beta/(3.0*alphac)*Hiso);
	      T = k*(1.0 - 3.0*k*Q*alphac) * itensor::dyadic(v1, v2);
            }
        }
    }
}




double splasticMP :: deviatoricEnergy() const
{
  istensor eps_e = istensor::deviatoricPart(eps_c - ep_c);
  return theElastoplasticMaterial.mu * eps_e.contract(eps_e);
}




void splasticMP :: deviatoricStress(istensor& s) const
{
  s = 2.0*theElastoplasticMaterial.mu*(istensor::deviatoricPart(eps_c-ep_c));
}




double splasticMP :: dissipatedEnergy() const
{
  const double mu    = theElastoplasticMaterial.mu;
  const double Hiso  = theElastoplasticMaterial.Hiso;
  const double Hkine = theElastoplasticMaterial.Hkine;
  const double k     = theElastoplasticMaterial.bulk;
    
    
  if (theElastoplasticMaterial.plasticityType=="mises")
    {
      const istensor ee_c = istensor::deviatoricPart(eps_c)-ep_c;
      const istensor s    = 2.0*mu*ee_c;
      const istensor Q    = -2.0/3.0 * Hkine*Xi_c;
      const double   q    = -Hiso*xi_c;
        
      const double dt_Psistar = (ep_c-ep_n).dot(s) + (Xi_c - Xi_n).dot(Q) + (xi_c-xi_n)*q;
      return dt_Psistar;
    }
    
  else if (theElastoplasticMaterial.plasticityType=="drucker")
    {
      const istensor e_cedev        = istensor::deviatoricPart(eps_c-ep_c);  // trial elastic deviatoric strain tensor
      const istensor e_cevol        = eps_c-ep_c-e_cedev;                    // trial elastic volumetric strain tensor
      const istensor stressdev      = 2.0 * mu * e_cedev;                     // deviatoric trial stress tensor
      const istensor stressvol      = 3.0 * k * e_cevol;                      // volumetric trial stress tensor
      istensor       sigma          = stressdev+stressvol;
      const double   q              = -Hiso*xi_c;
            
      const double dt_Psistar = (ep_c-ep_n).dot(sigma) + (xi_c-xi_n)*q;
      return dt_Psistar;
    }
  else
    {
      return 0.0;
    }
}




// when trial point lies beyond the apex, this return mapping must be used
// return value is the increment in volumetric plastic strain Delta-eps-p-vol
// dpvol is 3*trace(eps^p)
double splasticMP :: druckerApexReturn(const splasticMaterial& m, const istensor& sigma, const double& q,
                                       istensor& sigmabar, double& qbar)
{
  const double   alpha = m.alphac;
  const double   Hiso  = m.Hiso;
  const double   k     = m.bulk;
  const double   Y0    = m.Y0;
  const double   beta  = (3.0*alpha+sqrt(3.0))/3.0;
    
  // starting point, trial state
  const double   Itrial = sigma.invariant1();
  double dpvol  = 0.0;
    
    
  // Only valid for linear hardening
  ivector b, c, d;
  d(0) = 9.0*alpha*beta*k/(Hiso*beta*beta+9.0*alpha*alpha*k);
  d(1) = Hiso*beta*beta/(Hiso*beta*beta+9.0*alpha*alpha*k);
  d(2) = -d(0);
    
  c(0) = d(1);
  c(1) = -Hiso*alpha*beta/(Hiso*beta*beta+9.0*alpha*alpha*k);
  c(2) = 9.0*alpha*alpha*k/(Hiso*beta*beta+9.0*alpha*alpha*k);
    
  double Ibar;
  Ibar = d(0)*Y0 + d(1)*Itrial + d(2)*q;
  qbar = c(0)*Y0 + c(1)*Itrial + c(2)*q;
    
  sigmabar = Ibar/3.0*istensor::identity();
  dpvol    = (Itrial-Ibar)/(9.0*k);
    
    
  return dpvol;
}




// mapping return when trial point lies within the normal to the smooth part
// of the cone
// Drucker-Prager close point projection
// return delta gamma
// Note: only valid for linear isotropic hardening
double splasticMP :: druckerConeProjection(const splasticMaterial& m, const istensor& sigma, const double& q)
{
  const double   mu     = m.mu;
  const double   alphac = m.alphac;
  const double   k      = m.bulk;
  const double   Hiso   = m.Hiso;
  const double   Y0     = m.Y0;
  const double   beta   = (3.0*alphac+sqrt(3.0))/3.0;
    
  const double   I      = sigma.invariant1();
  const istensor s      = istensor::deviatoricPart(sigma);
  const double   J      = sqrt(s.J2());
    
  double dg = (J + alphac*I - beta*Y0 + beta*q)/(mu+9.0*alphac*alphac*k+beta*beta*Hiso);
    
  return dg;
}




// this function updates the current plastic variables for a given strain
void splasticMP :: druckerReturn(const istensor& strain, double& dg)
{
  const double mu     = theElastoplasticMaterial.mu;
  const double k      = theElastoplasticMaterial.bulk;
  const double alphac = theElastoplasticMaterial.alphac;
  const double beta   = (3.0*alphac+sqrt(3.0))/3.0;
  const double Hiso   = theElastoplasticMaterial.Hiso;
  const double Y0     = theElastoplasticMaterial.Y0;
    
  // trial state
  istensor ee_trial    = istensor::deviatoricPart(strain-ep_n);  // trial elastic deviatoric strain tensor
  double   te_trial    = strain.trace() - ep_n.trace();
  istensor s_trial     = 2.0 * mu * ee_trial;                    // deviatoric trial stress tensor
  istensor sigma_trial = s_trial + k* te_trial * istensor::identity();
  double   q_trial     = -Hiso*xi_n;
  istensor Q_trial;
  istensor sigma;
  double q;

    
  //check yield condition
  const double f_trial = yieldfunction(theElastoplasticMaterial, sigma_trial, Q_trial, q_trial);
    
  // compute stress and update internal variables
  if (f_trial <= PROJTOL*Y0)
    {
      // elastic step
        
      ep_c = ep_n;
      xi_c = xi_n;
      dg   = 0.0;
        
    }
    
  else  // plastic step: do return mapping
    {
      //closest point projection algorithm
      dg = druckerConeProjection(theElastoplasticMaterial, sigma_trial, q_trial);
      double  J_trial = sqrt(s_trial.J2());
        
      if (J_trial-mu*dg >= 0.0)
        {
	  //correct internal variables
	  ep_c   = ep_n + dg*s_trial/(2.0*J_trial) + dg*alphac * istensor::identity();
	  xi_c   = (Hiso > 0.0) ? xi_n + beta*dg : 0.0;
	  toApex = false;
        }
      else
        {
	  //apex dpvol is alphac*Delta gamma
            
	  toApex = true;
	  double dpvol = druckerApexReturn(theElastoplasticMaterial, sigma_trial, q_trial, sigma, q);
	  ep_c  = ep_n + dpvol* (istensor::identity()) + ee_trial;
	  xi_c  = xi_n + dpvol/alphac*beta;
        }
    }
    
  //Compute the deviatoric and volumetric elastic strain tensor in order to obtain
  //sigma for "assert" function
  istensor ee_c, epsve_c;
  ee_c    = istensor::deviatoricPart(strain-ep_c); //current elastic deviatoric strain tensor
  epsve_c = strain-ep_c-ee_c;                   //current elastic volumetric strain tensor
    
  const double   ytest = yieldfunction(theElastoplasticMaterial,
				       2.0*mu*ee_c+3.0*k*epsve_c, Q_trial, -Hiso*xi_c);
    
  assert(ytest/Y0 < 1e-3 );
}




materialState splasticMP :: getCurrentState() const
{
  materialState state_c = smallStrainMP::getCurrentState();

  state_c.theDouble.push_back(dg_c);
  state_c.theDouble.push_back(xi_c);
  state_c.theStensor.push_back(ep_c);
  state_c.theStensor.push_back(Xi_c);

  return state_c;
}




materialState splasticMP :: getConvergedState() const
{
  materialState state_n = smallStrainMP::getConvergedState();

  state_n.theDouble.push_back(dg_n);
  state_n.theDouble.push_back(xi_n);
  state_n.theStensor.push_back(ep_n);
  state_n.theStensor.push_back(Xi_n);

  return state_n;
}




double splasticMP :: plasticSlip() const
{
  return ep_c.norm();
}




void splasticMP :: onedReturn(const double& strain11, double& dg)
{
  const double E     = theElastoplasticMaterial.E;
  const double Hiso  = theElastoplasticMaterial.Hiso;
  const double Hkine = theElastoplasticMaterial.Hkine;
    
    
  // trial state
  const double   sigmatrial     = E*(strain11 - ep_n(0,0));
  const double   qtrial         = Hiso*xi_n;
  const double   Qtrial         = Hkine*Xi_n(0,0);
    
  //check yield condition
  const double   ftrial         = yieldfunction(theElastoplasticMaterial, sigmatrial, Qtrial, qtrial);
    
  // compute stress and update internal variables
  if (ftrial <= 1.0e-8)
    {
      //elastic step
      ep_c(0,0) = ep_n(0,0);
      Xi_c(0,0) = Xi_n(0,0);
      xi_c      = xi_n;
      dg        = 0.0;
    }
    
  else
    {
      // plastic slip
      dg = ftrial  / (E+Hiso+Hkine);
      // correct internal variables
      if((sigmatrial - Qtrial) >= 0.0)
        {
	  ep_c(0,0)  = ep_n(0,0) + dg;
	  Xi_c(0,0)  = Xi_n(0,0) + dg;
	  xi_c       = xi_n      + dg;
        }
      else
        {
	  ep_c(0,0)  = ep_n(0,0) - dg;
	  Xi_c(0,0)  = Xi_n(0,0) - dg;
	  xi_c       = xi_n      + dg;
        }
    }
}




// return to yield surface for von Mises model
void splasticMP :: radialReturn(const istensor& strain, double& dg)
{
  const double mu    = theElastoplasticMaterial.mu;
  const double Hiso  = theElastoplasticMaterial.Hiso;
  const double Hkine = theElastoplasticMaterial.Hkine;
    
    
  // trial state
  const istensor e_c            = istensor::deviatoricPart(strain);       // current deviatoric strain
  const istensor ee_c           = istensor::deviatoricPart(strain-ep_n);  // current elastic deviatoric strain
  const istensor strial         = 2.0 * mu * ee_c;                        // deviatoric trial stress
    
  const istensor Qtrial         = -2.0/3.0*Hkine*Xi_n;
  const double   qtrial         = -Hiso*xi_n;
    
  //check yield condition
  const double   ftrial = yieldfunction(theElastoplasticMaterial, strial, Qtrial, qtrial);
    
  // compute stress and update internal variables
  if (ftrial <= 1.0e-8)
    {
      // elastic step
      ep_c = ep_n;
      Xi_c = Xi_n;
      xi_c = xi_n;
      dg   = 0.0;
    }
    
  else
    {
      // plastic step: do return mapping
      const istensor backs = strial - Qtrial;
      const istensor n     = backs / backs.norm();
        
      // plastic slip
      dg = ftrial  / (2.0*mu + 2.0/3.0*(Hiso+Hkine));
        
      // correct internal variables
      ep_c  = ep_n + dg*n;
      Xi_c  = Xi_n - dg*n;
      xi_c  = xi_n + sqrt(2.0/3.0)*dg;
    }
    
  assert(yieldfunction(theElastoplasticMaterial, 2.0*mu*(e_c-ep_c), -2.0/3.0*Hkine*Xi_c, -Hiso*xi_c) < 1.0e-4);
}




void splasticMP :: resetCurrentState()
{
  smallStrainMP::resetCurrentState();

  dg_c   = dg_n;
  ep_c   = ep_n;
  Xi_c   = Xi_n;
  xi_c   = xi_n;
}




void splasticMP :: setConvergedState(const double theTime, const istensor & strainn,
                                     const istensor &epn, const double xin, const istensor &Xin)
{
  time_n    = theTime;
  ep_n      = epn;
  xi_n      = xin;
  eps_n     = strainn;
  Xi_n      = Xin;
}




void splasticMP :: setRandom()
{
  istensor tmp;
  tmp.setRandom();
    
  if (theElastoplasticMaterial.plasticityType == "mises")
    {
      ep_n = istensor::deviatoricPart(tmp);
      ep_c = ep_n;
        
      Xi_n = -ep_n;
      Xi_c = Xi_n;
        
      xi_n = muesli::randomUniform(1.0, 2.0);
      xi_c = xi_n;
    }
    
  else if (theElastoplasticMaterial.plasticityType == "drucker")
    {
      ep_n = istensor::symmetricPartOf(tmp);
      ep_c = ep_n;
        
      xi_n = muesli::randomUniform(1.0, 2.0);
      xi_c = xi_n;
    }
}




double splasticMP :: storedEnergy() const
{
  const double mu    = theElastoplasticMaterial.mu;
  const double Hiso  = theElastoplasticMaterial.Hiso;
  const double Hkine = theElastoplasticMaterial.Hkine;
  const double k     = theElastoplasticMaterial.bulk;
    
  double th = eps_c.trace() - ep_c.trace();
  double We = mu*(istensor::deviatoricPart(eps_c-ep_c)).squaredNorm() + 0.5 * k * th*th;
  double Wp = 0.5*Hiso*xi_c*xi_c + 0.5*2.0/3.0*Hkine*Xi_c.squaredNorm();
    
  return Wp+We;
}




void splasticMP :: stress(istensor& sigma) const
{
  if (theElastoplasticMaterial.plasticityType=="mises")
    {
      deviatoricStress(sigma);
      sigma += theElastoplasticMaterial.bulk*eps_c.trace()*istensor::identity();
    }
    
  else if (theElastoplasticMaterial.plasticityType=="drucker")
    {
      deviatoricStress(sigma);
      sigma += theElastoplasticMaterial.bulk*(eps_c-ep_c).trace()*istensor::identity();
    }
}




/*
  Voigt convention
 
  |1 6 5|
  |  2 4|
  |    3|
*/
void splasticMP :: stressVector(double S[6]) const
{
  const double nu = theElastoplasticMaterial.nu;
  const double k  = theElastoplasticMaterial.E/(1.0+nu)/(1.0-2.0*nu);
    
  S[0] = k*(1.0-nu) * (eps_c(0,0)-ep_c(0,0)) + k*nu * (eps_c(1,1)-ep_c(1,1)) + k*nu * (eps_c(2,2)-ep_c(2,2));
  S[1] = k*(1.0-nu) * (eps_c(1,1)-ep_c(1,1)) + k*nu * (eps_c(2,2)-ep_c(2,2)) + k*nu * (eps_c(0,0)-ep_c(0,0));
  S[2] = k*(1.0-nu) * (eps_c(2,2)-ep_c(2,2)) + k*nu * (eps_c(0,0)-ep_c(0,0)) + k*nu * (eps_c(1,1)-ep_c(1,1));
  S[3] = k*(1.0-2.0*nu)* (eps_c(1,2)-ep_c(1,2));
  S[4] = k*(1.0-2.0*nu)* (eps_c(2,0)-ep_c(2,0));
  S[5] = k*(1.0-2.0*nu)* (eps_c(0,1)-ep_c(0,1));
}




void splasticMP :: tangentTensor(itensor4& C) const //add the remaining J2 tangent
{
  C.setZero();

  if (theElastoplasticMaterial.plasticityType=="drucker")
    {
      if (dg_c == 0.0)
        {
	  for (unsigned i=0; i<3; i++)
	    for (unsigned j=0; j<3; j++)
	      for (unsigned k=0; k<3; k++)
		for (unsigned l=0; l<3; l++)
		  {
		    if (i==j && k==l) C(i,j,k,l) += theElastoplasticMaterial.lambda;
		    if (i==k && j==l) C(i,j,k,l) += theElastoplasticMaterial.mu;
		    if (i==l && j==k) C(i,j,k,l) += theElastoplasticMaterial.mu;
		  }
        }
      else
        {
	  const double   mu          = theElastoplasticMaterial.mu;
	  const double   kb          = theElastoplasticMaterial.bulk;
	  const double   Hiso        = theElastoplasticMaterial.Hiso;
	  const double   alphac      = theElastoplasticMaterial.alphac;
	  const istensor edtrial     = istensor::deviatoricPart(eps_c-ep_n);//deviatoric elastic trial strain
	  const istensor D           = edtrial * (1.0/edtrial.norm());
	  const double   normed      = sqrt(2.0)*edtrial.norm();
	  const double   root2       = sqrt(2.0);
	  const double   beta        = (3.0*alphac+sqrt(3.0))/3.0;
	  const double   A           = 1.0/(mu+9.0*kb*alphac*alphac+beta*beta*Hiso);
	  const istensor s_trial     = 2.0 * mu * edtrial;
	  const double   J_trial     = sqrt(s_trial.J2());
	  const double   Q           = 1.0/(3.0*kb*alphac+beta*beta/(3.0*alphac)*Hiso);
            

	  bool isInCone = J_trial-mu*dg_c >= 0.0;
	  if (isInCone)//case within cone's definition
            {
	      for (unsigned i=0; i<3; i++)
		for (unsigned j=0; j<3; j++)
		  for (unsigned k=0; k<3; k++)
		    for (unsigned l=0; l<3; l++)
		      {
			if (i==j && k==l) C(i,j,k,l) += -2.0*mu*(1.0-dg_c/normed)/3.0+kb-9.0*kb*kb*alphac*alphac*A;
			if (i==k && j==l) C(i,j,k,l) += mu*(1.0-dg_c/normed);
			if (i==l && j==k) C(i,j,k,l) += mu*(1.0-dg_c/normed);
			if (l==k)         C(i,j,k,l) += - A*2.0/root2*mu*3.0*kb*alphac*D(i,j);
			if (i==j)         C(i,j,k,l) += - A*2.0/root2*mu*3.0*kb*alphac*D(k,l);
			C(i,j,k,l) += (2.0*mu*(dg_c/normed)-2.0*mu*mu*A)*D(i,j)*D(k,l);
		      }
            }
	  else//case returned to the apex
            {
	      for (unsigned i=0; i<3; i++)
		for (unsigned j=0; j<3; j++)
		  for (unsigned k=0; k<3; k++)
		    for (unsigned l=0; l<3; l++)
		      {
			if (i==j && k==l) C(i,j,k,l) += kb-3.0*kb*kb*Q*alphac;
		      }
            }
        }
    }
  else if (theElastoplasticMaterial.plasticityType=="mises")
    {
      if (dg_c == 0.0)
        {
	  for (unsigned i=0; i<3; i++)
	    for (unsigned j=0; j<3; j++)
	      for (unsigned k=0; k<3; k++)
		for (unsigned l=0; l<3; l++)
		  {
		    if (i==j && k==l) C(i,j,k,l) += theElastoplasticMaterial.lambda;
		    if (i==k && j==l) C(i,j,k,l) += theElastoplasticMaterial.mu;
		    if (i==l && j==k) C(i,j,k,l) += theElastoplasticMaterial.mu;
		  }
        }
      else
        {
	  const double   mu     = theElastoplasticMaterial.mu;
	  const double   kb     = theElastoplasticMaterial.bulk;
	  const double   Hkine  = theElastoplasticMaterial.Hkine;
	  const double   Hiso   = theElastoplasticMaterial.Hiso;
	  const istensor strial = 2.0*mu*(istensor::deviatoricPart(eps_c) - ep_n);
	  const istensor Qtrial = -2.0/3.0*Hkine*Xi_n;
	  const istensor backs  = strial - Qtrial;
	  const istensor n      = backs * (1.0/backs.norm());
	  const double   zeta   = 1.0 - 2.0*mu*dg_c/backs.norm();
	  const double   zbar   = 2.0*mu / (2.0*mu+2.0/3.0*(Hiso+Hkine)) - (1.0-zeta);

	  for (unsigned i=0; i<3; i++)
	    for (unsigned j=0; j<3; j++)
	      for (unsigned k=0; k<3; k++)
		for (unsigned l=0; l<3; l++)
		  {
		    if (i==j && k==l) C(i,j,k,l) += kb-2.0*mu*zeta/3.0;
		    if (i==k && j==l) C(i,j,k,l) += mu*zeta;
		    if (i==l && j==k) C(i,j,k,l) += mu*zeta;
		    C(i,j,k,l) += -2.0*mu*zbar*n(i,j)*n(k,l);
		  }
        }
    }
}


// this function updates the current plastic variables for a given strain
void splasticMP :: trescaReturn(const istensor& strain, double& dg)
{
  const double mu     = theElastoplasticMaterial.mu;
  const double k      = theElastoplasticMaterial.bulk;
  const double Hiso   = theElastoplasticMaterial.Hiso;
  const double Y0     = theElastoplasticMaterial.Y0;

  // trial state
  istensor ee_trial    = istensor::deviatoricPart(strain-ep_n);  // trial elastic deviatoric strain tensor
  double   te_trial    = strain.trace();
  istensor s_trial     = 2.0 * mu * ee_trial;                    // deviatoric trial stress tensor
  istensor sigma_trial = s_trial + k * te_trial * istensor::identity();
  double   q_trial     = -Hiso*xi_n;
  istensor sigma;
  istensor Qdummy;

  //check yield condition
  double f_trial = yieldfunction(theElastoplasticMaterial, sigma_trial, Qdummy, q_trial);

  // compute stress and update internal variables
  if (f_trial <= PROJTOL*Y0)
    {
      // elastic step
      ep_c = ep_n;
      xi_c = xi_n;
      dg   = 0.0;
    }

  else  // plastic step: do return mapping
    {
      ivector sigtr, sig;
      ivector evec[3];
      sigma_trial.spectralDecomposition(evec, sigtr);

      //
      while (true)
        {
	  dg += f_trial/(4.0*mu+Hiso);
	  f_trial = sigtr(2) - sigtr(0) - 4.0*mu*dg - (Y0+Hiso*(xi_n+dg));
	  if (f_trial <= PROJTOL*Y0)
            {
	      sig(2) = sigtr(2) - 2.0*mu*dg;
	      sig(1) = sigtr(1);
	      sig(0) = sigtr(0) + 2.0*mu*dg;
	      xi_c    = xi_n + dg;
	      break;
            }
        }

      if (sig(0) > sig(1) || sig(1) > sig(2))
        {
	  if (sigtr(0) + sigtr(2) - 2.0*sigtr(1) > 0.0)
            {
	      // return to right corner
	      double dga(0.0), dgb(0.0);
	      double sa = sigtr(2) - sigtr(0);
	      double sb = sigtr(2) - sigtr(1);
	      double res[2];


	      res[0] = sa - (Y0+Hiso*(xi_n+dg));
	      res[1] = sb - (Y0+Hiso*(xi_n+dg));

	      double dgbar = dga + dgb;
	      xi_c = xi_n + dgbar;

	      double detH = (4.0*mu+Hiso)*(4.0*mu+Hiso)-(2.0*mu+Hiso)*(2.0*mu+Hiso);
	      double Hinv[2][2];
	      Hinv[0][0] = Hinv[1][1] = (-4.0*mu-Hiso)/detH;
	      Hinv[1][0] = Hinv[0][1] = (2.0*mu+Hiso)/detH;

	      dga -= Hinv[0][0]*res[0] + Hinv[0][1]*res[1];
	      dgb -= Hinv[1][0]*res[0] + Hinv[1][1]*res[1];
            }
	  else
            {
	      // return to left corner
            }

        }

      xi_c  = xi_n + dg;
      eps_c = istensor::deviatoricPart(strain);
      for (unsigned a=0; a<3; a++) eps_c.addScaledVdyadicV(-sig(a)/(2.0*mu), evec[a]);

    }

  //Compute the deviatoric and volumetric elastic strain tensor in order to obtain
  //sigma for "assert" function
  istensor ee_c, epsve_c;
  ee_c    = istensor::deviatoricPart(strain-ep_c); //current elastic deviatoric strain tensor
  epsve_c = strain-ep_c-ee_c;                   //current elastic volumetric strain tensor

  const double   ytest = yieldfunction(theElastoplasticMaterial,
				       2.0*mu*ee_c+3.0*k*epsve_c, Qdummy, -Hiso*xi_c);

  assert(ytest/Y0 < 1e-3);
}




void splasticMP :: updateCurrentState(const double theTime, const double strain11)//1D model
{
  smallStrainMP::updateCurrentState(theTime, strain11);
  onedReturn(strain11,dg_c);
}




void splasticMP :: updateCurrentState(const double theTime, const istensor& strain)
{
  smallStrainMP::updateCurrentState(theTime, strain);

  if (theElastoplasticMaterial.plasticityType == "mises")
    {
      radialReturn(strain, dg_c);
    }
    
  else if (theElastoplasticMaterial.plasticityType == "drucker")
    {
      druckerReturn(strain, dg_c);
    }

  else if (theElastoplasticMaterial.plasticityType == "tresca")
    {
      trescaReturn(strain, dg_c);
    }
    
  eps_c  = strain;
}




double splasticMP :: uniaxialStress() const
{
  double  E   = theElastoplasticMaterial.E;
  return  E*(eps_c(0,0) - ep_c(0,0));
}




double splasticMP :: uniaxialStiffness() const
{
    
  return 0.0;
}



double splasticMP :: volumetricStiffness() const
{
  double ret = 0.0;
  double k   = theElastoplasticMaterial.bulk;
  if (theElastoplasticMaterial.plasticityType == "mises")
    {
        
      ret = k;
    }
    
  if (theElastoplasticMaterial.plasticityType == "drucker")
    {
      const double alphac = theElastoplasticMaterial.alphac;
      const double mu     = theElastoplasticMaterial.mu;
      const double beta   = (3.0*alphac+sqrt(3.0))/3.0;
      const double Hiso   = theElastoplasticMaterial.Hiso;
      const double A      = 1.0/(mu+9.0*k*alphac*alphac+beta*beta*Hiso);
      const double Q      = 1.0/(3.0*k*alphac+beta*beta/(3.0*alphac)*Hiso);
      if (dg_c==0.0)
	ret = k;
      else
        {
	  const           istensor e_cedev        = istensor::deviatoricPart(eps_c-ep_n);
	  const           istensor stressdevtrial = 2.0 * mu * e_cedev;
	  const double    J_trial                 = sqrt(stressdevtrial.J2());
            
	  if (J_trial-mu*dg_c >= 0.0)
            {
	      ret = k-9.0*k*k*alphac*alphac*A;
            }
	  else
            {
	      ret = (1.0-3.0*k*Q*alphac)*k;
            }
        }
    }
    
  return ret;
}




double splasticMP :: volumetricEnergy() const
{
  double th = (eps_c-ep_c).trace();
  double V  = 0.5*theElastoplasticMaterial.bulk * th * th;
  return V;
}




double splasticMP :: yieldfunction(const splasticMaterial& m, const istensor& sigma, const istensor& Q, const double& q)
{
  const istensor s = istensor::deviatoricPart(sigma);
  const double   Y0 = m.Y0;
    
  if (m.plasticityType == "drucker")
    {
      const double   I      = sigma.invariant1();
      const double   J      = sqrt(s.J2());
      const double   alphac = m.alphac;
	       
      return J + alphac*I - (3.0*alphac+sqrt(3.0))/3.0*(Y0-q);
    }
    
  else if (m.plasticityType == "mises")
    {
      istensor backs = s - Q;
      return backs.norm() - sqrt(2.0/3.0)*(Y0 - q);
    }

  else if (m.plasticityType == "tresca")
    {
      ivector eval = sigma.eigenvalues();
      return eval(2) - eval(0) - (Y0 - q);
    }

  else
    {
      return 0.0;
    }
}




double splasticMP :: yieldfunction(const splasticMaterial& m, const double& sigma, const double& Q, const double& q)
{
  const double   Y0   = m.Y0;
  return abs(sigma - Q) - (Y0 + q);
}

