/****************************************************************************
 *
 *                                 M U E S L I   v 1.0
 *
 *
 *     Copyright 2016 IMDEA Materials Institute, Getafe, Madrid, Spain
 *     Contact: muesli.materials@imdea.org
 *     Author: Ignacio Romero (ignacio.romero@imdea.org)
 *
 *     This file is part of MUESLI.
 *
 *     MUESLI is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     MUESLI is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with MUESLI.  If not, see <http://www.gnu.org/licenses/>.
 *
*****************************************************************************/


#include <string.h>
#include <cmath>
#include "viscoplastic.h"
#include <cassert>

/*
 * viscoplastic.cpp,
 * infinitesimal kinematics visco-elastoplastic element
 * d. del pozo, i. romero, feb 2015
 *
 *
 *  1) von Mises Model with linear isotropic and kinematic hardenings
 *
 *     See "Numerical analysis and simulation of plasticity", by J.C. Simo,
 *     Handbook of numerical analysis vol VI
 *     table 18.2 and pages close to it
 *
 *     We = kappa/2 (theta)^2 + mu * ||e_elastic||^2
 *     Wp = Hiso/2 xi^2 + 2/3 Hkine/2 ||Xi||^2
 *     f(sigma, q, Q) = || dev(sigma) - Q || - sqrt(2/3) (Y0 - q)
 */


#define PROJTOL     1e-8     // residual tolerance for return-to-yield-surface solutions
#define PROJMAXITER 25      // max num of allowed interations in return-to-yield solutions


using namespace std;
using namespace muesli;





viscoplasticMaterial :: viscoplasticMaterial(const std::string& name,
                                             const materialProperties& cl)
:
smallStrainMaterial(name, cl),
plasticityType("mises"),
E(0.0), nu(0.0), bulk(0.0), cp(0.0), cs(0.0), lambda(0.0), mu(0.0), rho(0.0),
Hiso(0.0), Hkine(0.0), Y0(0.0),eta(0.0)
{
    muesli::assignValue(cl, "young",   E);
    muesli::assignValue(cl, "poisson", nu);
    muesli::assignValue(cl, "lambda",  lambda);
    muesli::assignValue(cl, "mu",      mu);
    muesli::assignValue(cl, "density", rho);
    muesli::assignValue(cl, "isotropich", Hiso);
    muesli::assignValue(cl, "kinematich", Hkine);
    muesli::assignValue(cl, "yieldstress", Y0);
    muesli::assignValue(cl, "eta", eta);

    if ( cl.find("model von_mises") != cl.end()) plasticityType = "mises";

    // intialize rest of constants
    if (lambda*lambda + mu*mu + E*E == 0)
        std::cout << "\n Error in small strain elastic material. No stiffness defined";

    // E and nu have priority. If they are defined, define lambda and mu
    if (E*E > 0.0)
    {
        lambda = E*nu/(1.0-2.0*nu)/(1.0+nu);
        mu     = E/2.0/(1.0+nu);
    }
    else
    {
        nu = lambda / 2.0 / (lambda+mu);
        E  = mu*2.0*(1.0+nu);
    }

    // we set all the constants, so that later on all of them can be recovered fast
    bulk = lambda + 2.0/3.0 * mu;

    if (rho > 0.0)
    {
        cp = sqrt((lambda+2.0*mu)/rho);
        cs = sqrt(2.0*mu/rho);
    }
}




viscoplasticMaterial :: viscoplasticMaterial(const double xE, const double xnu,
                                             const double xrho, const double xHiso,
                                             const double xHkine, const double xY0,
                                             const std::string xplst, const double xeta)
:
plasticityType(xplst),
E(xE), nu(xnu), bulk(0.0), cp(0.0), cs(0.0),
lambda(0.0), mu(0.0),
rho(xrho), Hiso(xHiso), Hkine (xHkine), Y0(xY0), eta(xeta)
{
    lambda = E*nu/(1.0-2.0*nu)/(1.0+nu);
    mu     = E/2.0/(1.0+nu);
    bulk   = lambda + 2.0/3.0 * mu;

    if (rho > 0.0)
    {
        cp   = sqrt((lambda+2.0*mu)/rho);
        cs   = sqrt(2.0*mu/rho);
    }
}




viscoplasticMaterial:: viscoplasticMaterial()
:
smallStrainMaterial(),
plasticityType("mises"),
E(0.0), nu(0.0), bulk(0.0), cp(0.0), cs(0.0), lambda(0.0), mu(0.0), rho(0.0),
Hiso(0.0), Hkine(0.0), Y0(0.0), eta(0.0)
{}




bool viscoplasticMaterial :: check() const
{
    if (mu > 0 && lambda+2.0*mu > 0) return true;
    else return false;
}




smallStrainMP* viscoplasticMaterial :: createMaterialPoint() const
{
    smallStrainMP* mp = new viscoplasticMP(*this);
    return mp;
}




double viscoplasticMaterial :: density() const
{
    return rho;
}




// this function is much faster than the one with string property names, because
// avoids string comparisons. It should be used.
double viscoplasticMaterial :: getProperty(const propertyName p) const
{
    double ret=0.0;

    switch (p)
    {
        case PR_LAMBDA:     ret = lambda;    break;
        case PR_MU:         ret = mu;        break;
        case PR_YOUNG:      ret = E;         break;
        case PR_POISSON:    ret = nu;        break;
        case PR_BULK:       ret = bulk;      break;
        case PR_CP:         ret = cp;        break;
        case PR_CS:         ret = cs;        break;
        case PR_YIELD:      ret = Y0;        break;
        case PR_ISOHARD:    ret = Hiso;      break;
        case PR_KINHARD:    ret = Hkine;     break;
        case PR_ETA    :    ret = eta;       break;


        default:
            std::cout << "Error in elastoplasticMaterial. Property not defined";
    }
    return ret;
}




void viscoplasticMaterial :: print(std::ostream &of) const
{

    if (plasticityType == "mises" )
    {
        of  << "\n Viscoelastoplastic material for small strain kinematics."
        << "\n   von Mises yield criterion with linear isotropic and kinematic hardenings."
        << "\n   Yield stress           : " << Y0
        << "\n   Isotropic hardening    : " << Hiso
        << "\n   Kinematic hardening    : " << Hkine
        << "\n   Viscosity              : " << eta;
    }


    of  << "\n   Young modulus:  E      : " << E;
    of  << "\n   Poisson ratio:  nu     : " << nu;
    of  << "\n   Lame constants: lambda : " << lambda;
    of  << "\n                   mu     : " << mu;
    of  << "\n   Bulk modulus:   k      : " << bulk;
    of  << "\n   Density                : " << rho;

    if (rho > 0.0)
    {
        of  << "\n   Wave velocities c_p    : " << cp;
        of  << "\n                   c_s    : " << cs;
    }
}




void viscoplasticMaterial :: setRandom()
{
    E   = muesli::randomUniform(1.0, 10000.0);
    nu  = muesli::randomUniform(0.05, 0.45);
    rho = muesli::randomUniform(1.0, 100.0);

    lambda = E*nu/(1.0-2.0*nu)/(1.0+nu);
    mu     = E/2.0/(1.0+nu);
    cp     = sqrt((lambda+2.0*mu)/rho);
    cs     = sqrt(2.0*mu/rho);
    bulk   = lambda + 2.0/3.0*mu;

    //Since in this first version only one viscoplastic material
    //has been included the setRandom() function has been bypassed with the code below ( int pt = 0; ), but is intended
    //to randomly pick any defined material between 0 and n -----> muesli::discreteUniform(0, n);

    int pt = 0;

    if (pt == 0)
    {
        plasticityType = "mises";
        Hiso  = E*1e-4 * muesli::randomUniform(1.0, 2.0);
        Hkine = E*1e-4 * muesli::randomUniform(1.0, 2.0);
        eta   = E*1e-5 * muesli::randomUniform(0.1, 1.0);
        Y0    = E*1e-3 * muesli::randomUniform(0.5, 1.5);
    }
}




bool viscoplasticMaterial :: test(std::ostream  &of)
{
    bool isok=true;
    setRandom();

    smallStrainMP* p = this->createMaterialPoint();
    p->setRandom();
    if      (this->plasticityType == "mises")   of << "\n   von Mises type";

    isok = p->testImplementation(of);

    return isok;
}




double viscoplasticMaterial :: waveVelocity() const
{
    return cp;
}




viscoplasticMP :: viscoplasticMP(const viscoplasticMaterial &m) :
muesli::smallStrainMP(m),
theViscoplasticMaterial(m),
dg_n(0.0), xi_n(0.0),
dg_c(0.0), xi_c(0.0)
{
    ep_n.setZero();
    ep_c.setZero();
    Xi_n.setZero();
    Xi_c.setZero();
}




void viscoplasticMP :: commitCurrentState()
{
    smallStrainMP::commitCurrentState();
    dg_n     = dg_c;
    ep_n     = ep_c;
    xi_n     = xi_c;
    Xi_n     = Xi_c;
}



/*
 void viscoplasticMP :: contractWithDeviatoricTangent(const ivector &v1, const ivector &v2, itensor &T) const
 {
 const double mu  = theViscoplasticMaterial.mu;
 const double eta = theViscoplasticMaterial.eta;
 const double tau = eta/(2.0*mu);
 const double dt  = time_c - time_n;

 if (dg_c > 0.0)
 {
 const double   Hiso     = theViscoplasticMaterial.Hiso;
 const double   Hkine    = theViscoplasticMaterial.Hkine;
 const istensor strial   = 2.0*mu*(istensor::deviatoricPart(eps_c) - ep_n);
 const istensor Qtrial   = -2.0/3.0*Hkine*Xi_n;

 const istensor backs    = strial - Qtrial;
 const double   betanorm = backs.norm();
 const istensor n        = backs * (1.0/betanorm);
 const double   zeta     = 1.0 - 2.0*mu*dg_c/betanorm;
 const double   zbar     = 2.0*mu/(1.0+tau/dt) / (2.0*mu+2.0/3.0*(Hiso+Hkine)) - (1.0-zeta);

 T = - 2.0/3.0*mu*zeta   * itensor::dyadic(v1, v2)
 +   mu*zeta*v1.dot(v2)  * itensor::identity()
 +   mu*zeta             * itensor::dyadic(v2,v1)
 -   2.0*mu*zbar         * itensor::dyadic(n*v1,n*v2);
 }
 else
 {
 T = - 2.0/3.0*mu        * itensor::dyadic(v1, v2)
 +   mu*v1.dot(v2)       * itensor::identity()
 +   mu                  * itensor::dyadic(v2,v1);
 }
 }
 */



void viscoplasticMP :: contractWithTangent(const ivector &v1, const ivector &v2, itensor &T) const
{
    const double mu  = theViscoplasticMaterial.mu;
    const double k   = theViscoplasticMaterial.bulk;
    const double eta = theViscoplasticMaterial.eta;
    const double Hkine = theViscoplasticMaterial.Hkine;
    const double Hiso = theViscoplasticMaterial.Hiso;
    const double tau = eta/(2.0*mu);
    const double dt  = time_c - time_n;

    if (theViscoplasticMaterial.plasticityType == "mises")
    {
        if (dg_c == 0.0)
        {
            T = (k - 2.0/3.0*mu) * itensor::dyadic(v1, v2)
            + mu*v1.dot(v2)      * itensor::identity()
            + mu                 * itensor::dyadic(v2,v1);
        }
        else
        {
            const istensor strial = 2.0*mu*(istensor::deviatoricPart(eps_c) - ep_n);
            const istensor Qtrial = -2.0/3.0*Hkine*Xi_n;
            const istensor backs  = strial - Qtrial;
            const istensor n      = backs * (1.0/backs.norm());
            const double zeta     = 1.0 - 2.0*mu*dg_c/backs.norm();
            const double zbar     = 2.0*mu/(1.0+tau/dt) / (2.0*mu+2.0/3.0*(Hiso+Hkine)) - (1.0-zeta);

            T = (k - 2.0/3.0*mu*zeta)  * itensor::dyadic(v1, v2)
            + (mu*zeta*v1.dot(v2)) * itensor::identity()
            + (mu*zeta)            * itensor::dyadic(v2,v1)
            - (2.0*mu*zbar)        * itensor::dyadic(n*v1,n*v2);
        }
    }
}




double viscoplasticMP :: deviatoricEnergy() const
{
    istensor eps_e = istensor::deviatoricPart(eps_c - ep_c);
    return theViscoplasticMaterial.mu * eps_e.contract(eps_e);
}




void viscoplasticMP :: deviatoricStress(istensor& s) const
{
    s = 2.0*theViscoplasticMaterial.mu*(istensor::deviatoricPart(eps_c)-ep_c);
}




double viscoplasticMP :: dissipatedEnergy() const
{
    const double mu    = theViscoplasticMaterial.mu;
    const double Hiso  = theViscoplasticMaterial.Hiso;
    const double Hkine = theViscoplasticMaterial.Hkine;

    if (theViscoplasticMaterial.plasticityType=="mises")
    {
        const istensor ee_c = istensor::deviatoricPart(eps_c)-ep_c;
        const istensor s    = 2.0*mu*ee_c;
        const istensor Q    = -2.0/3.0 * Hkine*Xi_c;
        const double   q    = -Hiso*xi_c;

        const double dt_Psistar = (ep_c-ep_n).dot(s) + (Xi_c - Xi_n).dot(Q) + (xi_c-xi_n)*q;
        return dt_Psistar;
    }

    else
    {
        return 0.0;
    }
}

materialState viscoplasticMP :: getConvergedState() const
{
    materialState state_n = smallStrainMP::getConvergedState();

    state_n.theDouble.push_back(dg_n);
    state_n.theDouble.push_back(xi_n);
    state_n.theStensor.push_back(ep_n);
    state_n.theStensor.push_back(Xi_n);

    return state_n;
}




materialState viscoplasticMP :: getCurrentState() const
{
    materialState state_c = smallStrainMP::getCurrentState();

    state_c.theDouble.push_back(dg_c);
    state_c.theDouble.push_back(xi_c);
    state_c.theStensor.push_back(ep_c);
    state_c.theStensor.push_back(Xi_c);

    return state_c;
}

double viscoplasticMP :: energyDissipationPotential() const
{
    const double mu    = theViscoplasticMaterial.mu;
    const double k     = theViscoplasticMaterial.bulk;
    const double Hiso  = theViscoplasticMaterial.Hiso;
    const double Hkine = theViscoplasticMaterial.Hkine;
    double ret   = 0.0;

    if (theViscoplasticMaterial.plasticityType == "mises")
    {
        istensor ee = istensor::deviatoricPart(eps_c)-ep_c;
        istensor s  = 2.0*mu*ee;
        istensor Q  = -2.0/3.0 * Hkine*Xi_c;
        double   q  = -Hiso*xi_c;
        double   t  = eps_c.trace();

        double We = mu*ee.squaredNorm() + 0.5*k*t*t;
        double Wp = 0.5*Hiso*xi_c*xi_c + 0.5*2.0/3.0*Hkine*Xi_c.squaredNorm();

        double dt_Psistar = (ep_c-ep_n).dot(s) + (Xi_c-Xi_n).dot(Q) + (xi_c-xi_n)*q;

        ret = We+Wp+dt_Psistar;
    }

    return ret;
}




double viscoplasticMP :: plasticSlip() const
{
    return ep_c.norm();
}




// return to yield surface for von Mises model
void viscoplasticMP :: radialReturn(const istensor& strain, double& dg)
{
    const double mu    = theViscoplasticMaterial.mu;
    const double Hiso  = theViscoplasticMaterial.Hiso;
    const double Hkine = theViscoplasticMaterial.Hkine;
    const double dt    = time_c - time_n;
    const double eta   = theViscoplasticMaterial.eta;
    const double tau   = eta/(2.0*mu);

    // trial state
    const istensor e_c            = istensor::deviatoricPart(strain);       // current deviatoric strain
    const istensor ee_c           = istensor::deviatoricPart(strain-ep_n);  // current elastic deviatoric strain
    const istensor strial         = 2.0 * mu * ee_c;                        // deviatoric trial stress

    const istensor Qtrial         = -2.0/3.0*Hkine*Xi_n;
    const double   qtrial         = -Hiso*xi_n;

    //check yield condition
    const double   ftrial = yieldfunction(theViscoplasticMaterial, strial, Qtrial, qtrial);

    // compute stress and update internal variables
    if (ftrial <= 1.0e-8)
    {
        // elastic step
        ep_c = ep_n;
        Xi_c = Xi_n;
        xi_c = xi_n;
        dg   = 0.0;
    }

    else
    {
        // plastic step: do return mapping

        const istensor backs = strial - Qtrial;
        const istensor n     = backs / backs.norm();

        // plastic slip
        dg = ftrial / ((1.0+tau/dt)*(2.0*mu + 2.0/3.0*(Hiso+Hkine)));

        // correct internal variables
        ep_c  = ep_n + dg*n;
        Xi_c  = Xi_n - dg*n;
        xi_c  = xi_n + sqrt(2.0/3.0)*dg;
    }
}




void viscoplasticMP :: resetCurrentState()
{
    smallStrainMP::resetCurrentState();
    dg_c   = dg_n;
    ep_c   = ep_n;
    Xi_c   = Xi_n;
    xi_c   = xi_n;
}




void viscoplasticMP :: setConvergedState(const double theTime, const double dgn, const istensor & epn,
                                         const double xin, const istensor & Xin, const istensor & strainn)
{
    time_n    = theTime;
    dg_n      = dgn;
    ep_n      = epn;
    xi_n      = xin;
    eps_n     = strainn;
    Xi_n      = Xin;
}




void viscoplasticMP :: setRandom()
{
    istensor tmp;
    tmp.setRandom();

    if (theViscoplasticMaterial.plasticityType == "mises")
    {
        ep_n = istensor::deviatoricPart(tmp);
        ep_c = ep_n;

        Xi_n = -ep_n;
        Xi_c = Xi_n;

        xi_n = muesli::randomUniform(1.0, 2.0);
        xi_c = xi_n;
    }

    else
    {
        cout << "\n Error initializating random viscoplastic material" << flush;
    }
}




double viscoplasticMP :: storedEnergy() const
{
    const double mu    = theViscoplasticMaterial.mu;
    const double Hiso  = theViscoplasticMaterial.Hiso;
    const double Hkine = theViscoplasticMaterial.Hkine;
    const double k     = theViscoplasticMaterial.bulk;

    double th = eps_c.trace() - ep_c.trace();
    double We = mu*(istensor::deviatoricPart(eps_c-ep_c)).squaredNorm() + 0.5 * k * th*th;
    double Wp = 0.5*Hiso*xi_c*xi_c + 0.5*2.0/3.0*Hkine*Xi_c.squaredNorm();

    return  Wp+We;
}




void viscoplasticMP :: stress(istensor& sigma) const
{
    if (theViscoplasticMaterial.plasticityType=="mises")
    {
        deviatoricStress(sigma);
        sigma += theViscoplasticMaterial.bulk*eps_c.trace()*istensor::identity();
    }
}


/*
 Voigt convention

 |1 6 5|
 |  2 4|
 |    3|

 */

void viscoplasticMP :: stressVector(double S[6]) const
{
    const double nu = theViscoplasticMaterial.nu;
    const double k  = theViscoplasticMaterial.E/(1.0+nu)/(1.0-2.0*nu);

    S[0] = k*(1.0-nu) * (eps_c(0,0)-ep_c(0,0)) + k*nu * (eps_c(1,1)-ep_c(1,1)) + k*nu * (eps_c(2,2)-ep_c(2,2));
    S[1] = k*(1.0-nu) * (eps_c(1,1)-ep_c(1,1)) + k*nu * (eps_c(2,2)-ep_c(2,2)) + k*nu * (eps_c(0,0)-ep_c(0,0));
    S[2] = k*(1.0-nu) * (eps_c(2,2)-ep_c(2,2)) + k*nu * (eps_c(0,0)-ep_c(0,0)) + k*nu * (eps_c(1,1)-ep_c(1,1));
    S[3] = k*(1.0-2.0*nu)* (eps_c(1,2)-ep_c(1,2));
    S[4] = k*(1.0-2.0*nu)* (eps_c(2,0)-ep_c(2,0));
    S[5] = k*(1.0-2.0*nu)* (eps_c(0,1)-ep_c(0,1));
}



void viscoplasticMP :: tangentTensor(itensor4& C) const //add the remaining J2 tangent
{
    C.setZero();


    if (theViscoplasticMaterial.plasticityType=="mises")
    {
        const double   mu     = theViscoplasticMaterial.mu;
        const double   kb     = theViscoplasticMaterial.bulk;
        const double   Hkine  = theViscoplasticMaterial.Hkine;
        const double   Hiso   = theViscoplasticMaterial.Hiso;
        const istensor strial = 2.0*mu*(istensor::deviatoricPart(eps_c) - ep_n);
        const istensor Qtrial = -2.0/3.0*Hkine*Xi_n;
        const istensor backs  = strial - Qtrial;
        const istensor n      = backs * (1.0/backs.norm());
        const double   zeta   = 1.0 - 2.0*mu*dg_c/backs.norm();
        const double   zbar   = 2.0*mu / (2.0*mu+2.0/3.0*(Hiso+Hkine)) - (1.0-zeta);


        if (dg_c == 0.0)
        {
            for (unsigned i=0; i<3; i++)
                for (unsigned j=0; j<3; j++)
                    for (unsigned k=0; k<3; k++)
                        for (unsigned l=0; l<3; l++)
                        {
                            if (i==j && k==l) C(i,j,k,l) += theViscoplasticMaterial.lambda;
                            if (i==k && j==l) C(i,j,k,l) += theViscoplasticMaterial.mu;
                            if (i==l && j==k) C(i,j,k,l) += theViscoplasticMaterial.mu;
                        }
        }
        else
        {
            for (unsigned i=0; i<3; i++)
                for (unsigned j=0; j<3; j++)
                    for (unsigned k=0; k<3; k++)
                        for (unsigned l=0; l<3; l++)
                        {
                            if (i==j && k==l) C(i,j,k,l) += kb-2.0*mu*zeta/3.0;
                            if (i==k && j==l) C(i,j,k,l) += mu*zeta;
                            if (i==l && j==k) C(i,j,k,l) += mu*zeta;
                            C(i,j,k,l) += -2.0*mu*zbar*n(i,j)*n(k,l);
                        }
        }
    }
}



void viscoplasticMP :: updateCurrentState(const double theTime, const double strain11)
{
    smallStrainMP::updateCurrentState(theTime, strain11);
}




void viscoplasticMP :: updateCurrentState(const double theTime, const istensor& strain)
{
    smallStrainMP::updateCurrentState(theTime, strain);
    
    if (theViscoplasticMaterial.plasticityType == "mises")
    {
        radialReturn(strain, dg_c);
        
    }
    
    eps_c = strain;
}




double viscoplasticMP :: uniaxialStress() const
{
    //Need to be defined
    return 0.0;
}




double viscoplasticMP :: uniaxialStiffness() const
{
    //Need to be defined
    return 0.0;
}




double viscoplasticMP :: volumetricEnergy() const
{
    double th = (eps_c-ep_c).trace();
    double V  = 0.5*theViscoplasticMaterial.bulk * th * th;
    return V;
}




double viscoplasticMP :: yieldfunction(const viscoplasticMaterial& m, const istensor& sigma, const istensor& Q, const double& q)
{
    const istensor s = istensor::deviatoricPart(sigma);
    const double   Y0 = m.Y0;
    
    if (m.plasticityType == "mises")
    {
        istensor backs = s - Q;
        return backs.norm() - sqrt(2.0/3.0)*(Y0 - q);
    }
    else
    {
        return 0.0;
    }
}
