/****************************************************************************
 *
 *                                 M U E S L I   v 1.0
 *
 *
 *     Copyright 2016 IMDEA Materials Institute, Getafe, Madrid, Spain
 *     Contact: muesli.materials@imdea.org
 *     Author: Ignacio Romero (ignacio.romero@imdea.org)
 *
 *     This file is part of MUESLI.
 *
 *     MUESLI is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     MUESLI is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with MUESLI.  If not, see <http://www.gnu.org/licenses/>.
 *
*****************************************************************************/



#include <stdio.h>
#include "smallthermo.h"
#include "muesli/Smallstrain/smallstrainlib.h"

using namespace muesli;

smallThermomechanicalMaterial :: smallThermomechanicalMaterial(const std::string& name,
                                                               const materialProperties& cl)
:
material(name, cl),
theSSMaterial(nullptr),
alpha(0.0), tref(0.0)
{
    if       (cl.find("type elastic") != cl.end())      theSSMaterial = new elasticMaterial(name, cl);
    else if  (cl.find("type splastic") != cl.end())     theSSMaterial = new splasticMaterial(name, cl);
    else if  (cl.find("type viscoelastic") != cl.end()) theSSMaterial = new viscoelasticMaterial(name, cl);
    else if  (cl.find("type viscoplastic") != cl.end()) theSSMaterial = new viscoplasticMaterial(name, cl);

    muesli::assignValue(cl, "alpha",   alpha);
    muesli::assignValue(cl, "tref",    tref);
}




smallThermomechanicalMaterial :: ~smallThermomechanicalMaterial()
{
    if (theSSMaterial != nullptr) delete(theSSMaterial);
}




smallThermomechanicalMP :: smallThermomechanicalMP(const smallThermomechanicalMaterial &m)
:
temp_n(0.0), temp_c(0.0),
theSSMP(nullptr),
theThermoMechMaterial(m)
{
    theSSMP = m.theSSMaterial->createMaterialPoint();
    gradt_n.setZero();
    gradt_c.setZero();
}




smallThermomechanicalMP :: ~smallThermomechanicalMP()
{
    if (theSSMP != nullptr) delete (theSSMP);
}




void  smallThermomechanicalMP :: commitCurrentState()
{
    temp_n = temp_c;
    theSSMP->commitCurrentState();
}




void smallThermomechanicalMP :: deviatoricStress(istensor& s) const
{
    theSSMP->deviatoricStress(s);
}




double smallThermomechanicalMP :: entropy() const
{
    const double k     = theSSMP->parentMaterial().getProperty(PR_BULK);
    const double beta  = 3.0 * k * theThermoMechMaterial.alpha;

    materialState state_c = theSSMP->getCurrentState();
    const istensor& eps_c = state_c.theStensor[0];

    return  beta * eps_c.trace();
}





double smallThermomechanicalMP :: freeEnergy() const
{
    const double k     = theSSMP->parentMaterial().getProperty(PR_BULK);
    const double beta  = 3.0 * k * theThermoMechMaterial.alpha;
    const double tref  = theThermoMechMaterial.tref;

    materialState state_c = theSSMP->getCurrentState();
    const istensor& eps_c = state_c.theStensor[0];


    double psi_mech = theSSMP->storedEnergy();
    double psi_coup = - beta * (temp_c - tref) * eps_c.trace();

    return psi_mech + psi_coup;
}




ivector smallThermomechanicalMP :: heatflux() const
{
    return -theThermoMechMaterial.conductivity * gradt_c;
}




double smallThermomechanicalMP :: plasticSlip() const
{
    return theSSMP->plasticSlip();
}




double smallThermomechanicalMP :: pressure() const
{
    return theSSMP->pressure();
}




void  smallThermomechanicalMP :: resetCurrentState()
{
    temp_c  = temp_n;
    gradt_c = gradt_n;
    theSSMP->resetCurrentState();
}




void smallThermomechanicalMP :: stress(istensor& sigma) const
{
    theSSMP->stress(sigma);
}




void smallThermomechanicalMP :: updateCurrentState(const double t, const double strain11, const double temp, const double gradt)
{
    temp_c     = temp;
    gradt_c(0) = gradt;
    theSSMP->updateCurrentState(t, strain11);
}




void smallThermomechanicalMP :: updateCurrentState(const double t, const istensor& strain, const double temp, const ivector& gradt)
{
    temp_c  = temp;
    gradt_c = gradt;
    theSSMP->updateCurrentState(t, strain);
}

